# Created by Alfred Ferrer Florensa
"""Contains objects to contain HIT"""
import os
import json
from cgelib.utils.file_mixin import CGELibFileError
from cgelib.output.result import Result
from cgelib.utils.file_mixin import ResultFile
from cgelib.utils.savers_mixin import StringCreator
from cgelib.sequence.Feature import CGEFeatureValueError, CGEFeatureFormatError, CGEFeatureKeyError, AlnFeature


class CGEHitTypeError(TypeError):
    """ Error related to a ValueError of the value of the feature
    """

    def __init__(self, message, *args):
        self.message = message
        # allow users initialize misc. arguments as any other builtin Error
        super(CGEHitTypeError, self).__init__(message, *args)


class CGEHitValueError(ValueError):
    """ Error related to a ValueError of the value of the feature
    """

    def __init__(self, message, *args):
        self.message = message
        # allow users initialize misc. arguments as any other builtin Error
        super(CGEHitValueError, self).__init__(message, *args)


class CGEHitKeyError(KeyError):
    """ Error related to a ValueError of the value of the feature
    """

    def __init__(self, message, *args):
        self.message = message
        # allow users initialize misc. arguments as any other builtin Error
        super(CGEHitKeyError, self).__init__(message, *args)


class AlnHit(dict):

    beone_json_path = Result.beone_json_path

    def __init__(self, software=None, empty=False, data=None,
                 fmt_file=beone_json_path, orig_file=None):

        """
            Root for all submodule of alignment
            Input:
                software: Software that has created the alignment
                file: File object from which the data has been extracted
                empty: If the query did hit on a template or did not
                data: Dictionary containing information about a single hit from
                      aligner

            Method creates a hit dict as defined in the BeOne template
            https://bitbucket.org/genomicepidemiology/cgelib/src/master/src/
            cgelib/output/templates_json/beone/
            from res.
        """
        self.software = software

        with open(fmt_file, "r") as fh:
            json_defs = json.load(fh)
            self.feature_defs = json_defs["aln_hit"]

        self.resultfiles = []
        self.file_type = []
        self.file_paths = []

        self["type"] = "aln_hit"

        if isinstance(orig_file, ResultFile):
            self.add_resultfile(orig_file)
        elif orig_file is None:
            self["template_file"]=None
        else:
            raise TypeError("The origin file cannot be %s " % (orig_file))

        self["templateID"] = None


        if data is not None:
            self.add_features(data)
        else:
            pass

        if isinstance(empty, bool):
            self.empty = empty
        else:
            raise TypeError("Empty variable has to be a boolean")


    def __str__(self):
        if self.software is None:
            aln_str = ""
        else:
            aln_str = self.software.title()
        if self.empty:
            string_hit = "No Hit by %s on %s" % (self["queryID"],
                                                 self["templateID"])
        else:
            if self["templateID"] is not None:
                string_hit = "%sHit on %s(\n" % (aln_str, self["templateID"])
                for key, value in self.items():
                    string_hit += "\t%s: %s,\n" % (key, value)
                string_hit = string_hit[:-2]
                string_hit += ")"
            elif self["queryID"] is not None:
                string_hit = "No Hit by %s" % self["queryID"]
            else:
                string_hit = "No Hit"
        return string_hit

    def __repr__(self):
        if self.software is None:
            aln_str = ""
        else:
            aln_str = self.software.title()
        if self["templateID"] is not None:
            string_hit = "%sHit on %s (" % (aln_str, self["templateID"])
            for key, value in self.items():
                string_hit += "%r, " % (value)
            string_hit = string_hit[:-1]
            string_hit += ")"
        else:
            string_hit = "No hit"
        return string_hit

    def add_features(self, data):
        if isinstance(data, dict):
            for key, value in data.items():
                self[key] = value
        else:
            raise CGEFeatureFormatError("Only an a dictionary"
                                        " of features with their values can be "
                                        "incorporated")

    def __getitem__(self, y):
        if y in self:
            return super(AlnHit, self).__getitem__(y)
        else:
            raise CGEHitKeyError("The feature %s has not been set for this hit."
                                  % (y))


    def __setitem__(self, i, y):

        if i in self and self[i] is not None:
            if y == self[i]:
                pass
            else:
                raise CGEFeatureValueError("Feature %s has been already "
                                           "attributed to the AlnHit with the "
                                           "value %s and is trying to be "
                                           "changed by the value %s." % (
                                            i, self[i], y))
        else:
            super(AlnHit, self).__setitem__(i, y)

    @staticmethod
    def dfs_to_dict(hit):
        for aln_hit in hit.file_type:
            if aln_hit == "Matrix":
                feature_df = hit.pop("matrix")
                feature_dict = AlnFeature.FeatureDF_to_dict(
                                feature_DF=feature_df, aln_hit_key=hit["key"],
                                resultfile=aln_hit)
                hit.update({"matrix_position":feature_dict})
            elif aln_hit == "Fragments":
                feature_df = hit.pop("reads_aligned")
                feature_dict = AlnFeature.FeatureDF_to_dict(
                                feature_DF=feature_df, aln_hit_key=hit["key"],
                                resultfile=aln_hit)
                hit.update({"fragment_aligned":feature_dict})
            elif aln_hit == "VCF":
                feature_df = hit.pop("point_variations")
                feature_dict = AlnFeature.FeatureDF_to_dict(
                                feature_DF=feature_df, aln_hit_key=hit["key"],
                                resultfile=aln_hit)
                hit.update({"point_variations":feature_dict})
        return hit

    @staticmethod
    def get_run_key(software_execs, hit):
        try:
            template_file = hit["template_file"]
        except KeyError:
            template_file = None
        if template_file is None:
            key_exec = None
        else:
            key_exec = None
            for k in software_execs:
                if k.endswith(template_file):
                    if key_exec is None:
                        key_exec = k
                    else:
                        raise ValueError("The run_key has not been able to set"
                                         " because several run keys coincide")
        return key_exec



    @staticmethod
    def get_rnd_unique_hit_key(hit_key, hit_collection,
                                minimum_hit_key, delimiter):
        """
            Input:
                hit_key: None-unique key
                hit_collection: Result object created by the cgelib package.
                minimum_key: Key prefix
                delimiter: String used as delimiter inside the returned key.
            Output:
                hit_key: Unique key (string)

            If hit_key is found in hit_collection. Creates a unique key by
            appending a random string ton minimum_hit_key.
        """
        while(hit_key in hit_collection):
            rnd_str = StringCreator.random_string(str_len=4)
            hit_key = ("{key}{deli}{rnd}"
                        .format(key=minimum_hit_key, deli=delimiter,
                                rnd=rnd_str))
        return hit_key


    def _get_unique_hit_key(self, hit_collection, delimiter=";;"):
        """
            Input:
                hit_collection: Result object created by the cgelib package.
                delimiter: String used as delimiter inside the returned key.
            Output:
                key: Unique key for hit

            Creates a unique key for Hit instance. Key format depends on
            database.
        """
        if self.empty:
            if self["template_file"] is not None:
                hit_key = "NoHit{deli}{query}{deli}on{deli}{template}".format(
                        deli=delimiter, query=self["queryID"],
                        template=self["template_file"])
            else:
                hit_key = "NoHit{deli}{query}".format(deli=delimiter,
                                                query=self["queryID"])
        else:
            if self["template_file"] is not None:
                hit_key = "{template}{deli}{templatefile}".format(
                        template=self["templateID"], deli=delimiter,
                        templatefile=self["template_file"])
            else:
                hit_key = "{template}".format(
                        template=self["templateID"])
        minimum_key = hit_key
        if hit_key in hit_collection:
            hit_key = AlnHit.get_rnd_unique_hit_key(
                        hit_key, hit_collection, minimum_key, delimiter)
        self["key"] = hit_key



class BlastnHit(AlnHit):
    """Object that is a hit on a database from BLAST"""

    def __init__(self, software="blast", data=None, empty=False,
                 orig_file=None):


        super().__init__(software=software, empty=empty, data=data,
                         orig_file=orig_file)

    def add_resultfile(self, resultfile):

        if isinstance(resultfile, list):
            raise CGEFeatureFormatError("Hit from blast cannot provide from"
                                        " different files")
        if(len(self.resultfiles) > 0 or len(self.file_paths) > 0 or
            len(self.file_type) > 0):
            raise CGEFeatureFormatError("This hit has already data from "
                                        "another file")
        else:
            self.resultfiles.append(resultfile)
            self.file_paths.append(resultfile.file_path)
            self.file_type.append(resultfile.type)

        try:
            self["template_file"] = resultfile.name
        except CGEFeatureValueError:
            raise CGELibFileError("When mixing hits from different resultfiles"
                                  ", these should come from the same alignment"
                                  ". The result files that are trying to be "
                                  "mixed (%s and %s) come from different "
                                  "alignments when adding the file %s." % (
                                  self["template_file"], result_file.name,
                                  result_file.type))


class KMAHit(AlnHit):
    """Object that is a hit on a database from KMA"""

    def __init__(self, software="kma", data=None, empty=False,
                 orig_file=None):


        super().__init__(software=software, empty=empty, data=data,
                         orig_file=orig_file)


    def add_hit(self, hit):
        hit_merged = KMAHit.merge_hits([self, hit])
        self.clear()
        self.file_type = hit_merged.file_type
        self.update(hit_merged)

    def add_resultfile(self, resultfile):

        if not isinstance(resultfile, list):
            resultfile_lst = [resultfile]
        else:
            resultfile_lst = resultfile

        for result_file in resultfile_lst:
            if result_file in self.resultfiles:
                raise CGELibFileError("The hits trying to be merged should "
                                      "come from different result files. The "
                                      "file %s has been already used for "
                                      "building this hit." % (result_file))
            self.resultfiles.append(result_file)
            self.file_paths.append(result_file.file_path)
            self.file_type.append(result_file.type)


            try:
                self["template_file"] = result_file.name
            except CGEFeatureValueError:
                raise CGELibFileError("When mixing hits from different resultfiles"
                                  ", these should come from the same alignment"
                                  ". The result files that are trying to be "
                                  "mixed (%s and %s) come from different "
                                  "alignments when adding the file %s." % (
                                  self["template_file"], result_file.name,
                                  result_file.type))


    @staticmethod
    def merge_hits(list_hits):
        merged_kmahit = list_hits.pop()
        while len(list_hits) > 0:
            hit = list_hits.pop()
            merged_kmahit.add_resultfile(hit.resultfiles)
            # Add data if templateID is same
            if("templateID" not in hit):
                raise CGEHitValueError(("The hit trying to be merged to the \
                                        existing hit is not defined by the \
                                        templateID"))
            elif(merged_kmahit["templateID"] is not None
                and str(merged_kmahit["templateID"]).rstrip() != str(hit["templateID"]).rstrip()):
                raise CGEHitValueError(("The two hits trying to merge do not \
                                         come from the same template (%r and \
                                         %r)" % (
                                         merged_kmahit["templateID"],
                                         hit["templateID"])))
            else:
                merged_kmahit.add_features(hit)
        return merged_kmahit



# Created by Alfred Ferrer Florensa
"""Contains objects for reading files"""

import os
import sys
import gzip
import signal
from cgelib.utils.format_mixin import FormatFile


class CGELibFileError(Exception):
    """ Raised when a file produced by CGELib has not been created and does not
        exists or is incompatible with other CGELib produced files.
    """

    def __init__(self, message, *args):
        self.message = message
        # allow users initialize misc. arguments as any other builtin Error
        super(CGELibFileError, self).__init__(message, *args)


class CGELibFileParseError(StopIteration):
    """ Raised when the iterator has arrived to the end.
    """

    def __init__(self, message, *args):
        self.message = message
        # allow users initialize misc. arguments as any other builtin Error
        super(CGELibFileParseError, self).__init__(message, *args)


class _File:
    """Generic interface for taking information from files"""

    def __init__(self, file_path=None, is_empty=None, compression=None,
                 extension=None):
        self.file_path = file_path
        self.is_empty = is_empty
        self.compression = compression
        if extension is None:
            split_extension = os.path.splitext(file_path)[1]
            self.extension = split_extension
        else:
            self.extension = extension

    def define_file(self, file_path=None):

        if file_path is not None:
            self.file_path = "%s%s" % (file_path, self.extension)

        if not os.path.isfile(self.file_path):
            raise OSError("File %s does not exists" % (self.file_path))

        self.compression = FormatFile.is_gzipped(self.file_path)

        if not os.access(self.file_path, os.R_OK):
            raise OSError("File's %s permissions do not allow to read it")

    def __str__(self):
        """TODO: Truncated files"""
        self.define_file()

        return self.file_path

    def __repr__(self):
        """Represent file as a string"""
        self.define_file()
        return "File(file=%s, compression=%s)" % (self.file_path,
                                                  self.compression)


class _Parse_File(object):
    """A generator that iterates through a CC-CEDICT formatted file, returning
    a tuple of parsed results (Traditional, Simplified, Pinyin, English)"""

    def __init__(self, path, is_gzip=False):
        self.path = path
        if is_gzip:
            self.file = gzip.open(path, "rt")
        else:
            self.file = open(path, "r")
        signal.signal(signal.SIGINT, self.signal_term_handler)

    def signal_term_handler(self, signal, frame):
        """ Make sure to close handle if app closes prematurely """
        try:
            self.close()
        except ValueError:
            pass
        sys.exit()

    def parse_line(self, file_type="line"):
        self.opened_file()
        if file_type=="xml":
            try:
                line = next(self.iterator_XML)
            except StopIteration:
                raise CGELibFileParseError(
                    """The iterator of the file %s has arrived to the end """
                    """of the file""" % self.path)
        else:
            try:
                line = self.file.readline()
            except ValueError:
                raise CGELibFileParseError(
                    """The iterator of the file %s has arrived to the end """
                    """of the file""" % self.path)
        return line

    def opened_file(self):
        if self.file.closed:
            raise CGELibFileParseError(
                """The iterator of the file %s has arrived to the end """
                """of the file""" % self.path)

    def close(self):
        self.file.close()


class ResultFile(_File):

    def __init__(
        self,
        type,
        file_path,
        name=None,
        read_method=None,
        extension=None,
        compression=None,
        translate_read=None,
        options_read=None,
    ):
        self.type = type
        if os.path.splitext(file_path)[-1] == "":
            if extension is None:
                raise ValueError("The file does not have an extension")
            else:
                resultfile_path = "%s%s" % (file_path, extension)
        else:
            resultfile_path = file_path

        if name is None:
            self.name = os.path.splitext(os.path.basename(file_path))[0]
        else:
            self.name = name
        self.read_method = read_method
        if options_read is None:
            self.options_read = {}
        else:
            if not isinstance(options_read, dict):
                raise TypeError("The parameters for a reading the results file"
                                " have to be in a dictionary")
            else:
                self.options_read = options_read
        _File.__init__(self, file_path=resultfile_path,
                       compression=compression, extension=extension)

    def __str__(self):
        return self.file_path

    def __repr__(self):
        return "%s(file=%s, compression=%s, read_method=%s)" % (
                self.type.capitalize(), self.file_path, self.compression,
                self.read_method)

    def read(self, **kwargs):

        _File.define_file(self)
        if self.read_method is None:
            read_function = None
        else:
            if bool(self.options_read):
                read_function = self.read_method(self.file_path,
                                                 **self.options_read)
            else:
                read_function = self.read_method(self.file_path)
        return read_function

    def dump_json(self):
        pass


class Iterator_BioFiles:

    BioFiles = ["fasta", "paired_end_reads", "single_end_reads"]

    def __init__(self, dir_paths, extension, type_file):

        if isinstance(dir_paths, list):
            self.dir_paths = iter(dir_paths)
        else:
            self.dir_paths = iter([dir_paths])

        if type_file not in Iterator_BioFiles.BioFiles:
            raise TypeError("The only types of data accepted by the class %s"
                            " are %s" % (self.__name__,
                                        ", ".join(Iterator_BioFiles.BioFiles)))
        self.extension = extension
        self.type_file = type_file
        self.folder_files = None
        self.path = None

    def __iter__(self):
        return self

    def __iter__(self):

        if self.folder is None:
            folder = next(self.dir_paths)
            if os.path.isdir(folder):
                self.folder_files = os.listdir(folder)
                self.path = folder
            else:
                raise OSError("The path %s is not a folder" % folder)
        check_files = True
        while check_files:
            if self.type_file:
                continue



# Created by Alfred Ferrer Florensa
"""Contains utils"""
import os
import json
import numpy as np
import pandas as pd


class FormatFile:

    @staticmethod
    def is_gzipped(file_path):
        ''' Returns True if file is gzipped and False otherwise.
            The result is inferred from the first two bits in the file read
            from the input path.
            On unix systems this should be: 1f 8b
            Theoretically there could be exceptions to this test but it is
            unlikely and impossible if the input files are otherwise expected
            to be encoded in utf-8.
        '''
        with open(file_path, mode='rb') as fh:
            bit_start = fh.read(2)
        if(bit_start == b'\x1f\x8b'):
            return True
        else:
            return False

    @staticmethod
    def get_input_format(input_files):
        """
        Takes all input files and checks their first character to assess
        the file format. Returns one of the following strings; fasta, fastq,
        other or mixed. fasta and fastq indicates that all input files are
        of the same format, either fasta or fastq. other indiates that all
        files are not fasta nor fastq files. mixed indicates that the inputfiles
        are a mix of different file formats.
        """

        # Open all input files and get the first character
        file_format = []
        invalid_files = []
        for infile in input_files:
            if is_gzipped(infile):
                f = gzip.open(infile, "rb")
                fst_char = f.read(1)
            else:
                f = open(infile, "rb")
                fst_char = f.read(1)
            f.close()
            # Assess the first character
            if fst_char == b"@":
                file_format.append("fastq")
            elif fst_char == b">":
                file_format.append("fasta")
            else:
                invalid_files.append("other")
        if len(set(file_format)) != 1:
            return "mixed"
        return ",".join(set(file_format))



class FormatVariable:

    @staticmethod
    def is_list_or_strings(lst):
        if isinstance(lst, str):
            return [lst]
        elif(bool(lst) and isinstance(lst, list)
                and all(isinstance(elem, str) for elem in lst)):
            return lst
        else:
            raise TypeError("The datasets variable have to be a string or a"
                            "list of strings")

    @staticmethod
    def is_a_list(lst):
        if isinstance(lst, list):
            list_lst = lst
        else:
            try:
                list_lst = list(lst)
            except ValueError:
                raise ValueError("Could not convert string to float: %s" % lst)
        return list_lst



import os
import json
import string
import random
import numpy as np
import pandas as pd


class Pd_NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, pd.DataFrame):
            return obj.to_dict()
        return super(Pd_NpEncoder, self).default(obj)


class SaveJson:
    @staticmethod
    def from_dict(object, file_path, overwrite=False):
        '''Save a dict into a json file'''
        # Check if folder exists
        if not os.path.isdir(os.path.dirname(file_path)):
            raise OSError("The folder where the json was going to be saved "
                          "do not exists")
        # Care about overwritting
        if overwrite is False and os.path.file(file_path):
            raise OSError("There is a file already called %s. If you want to"
                          " overwrite that file, set overwrite to True.")

        with open(file_path, 'w') as fp:
            json.dump(object, fp, indent=4, cls=Pd_NpEncoder)

    @staticmethod
    def dump_json(std_result_file, std_result):
        with open(std_result_file, 'w') as fh:
            fh.write(std_result.json_dumps())


class StringCreator:

    @staticmethod
    def random_string(str_len=4):
        """
            Output:
                random string of length 'str_len'

            Return a random string of the provided length. The string will only
            consist of lowercase ascii letters.
        """
        letters = string.ascii_lowercase
        return ''.join(random.choice(letters) for i in range(str_len))

    @staticmethod
    def text_table(headers, rows, empty_replace='-'):
        ''' Create text table

        USAGE:
            >>> from tabulate import tabulate
            >>> headers = ['A','B']
            >>> rows = [[1,2],[3,4]]
            >>> print(text_table(headers, rows))
            **********
              A     B
            **********
              1     2
              3     4
            ==========
        '''
        # Replace empty cells with placeholder
        rows = map(lambda row: map(lambda x: x if x else empty_replace, row),
                   rows)
        # Create table
        table = tabulate(rows, headers, tablefmt='simple').split('\n')
        # Prepare title injection
        width = len(table[0])
        # Switch horisontal line
        table[1] = '*' * (width + 2)
        # Update table with title
        table = (("%s\n" * 3)
                 % ('*' * (width + 2), '\n'.join(table), '=' * (width + 2)))
        return table



    # Created by Alfred Ferrer Florensa
"""Contains objects for reading """

from cgelib.alignment.KMA.alignment_files import Iterator_ResFile, Iterator_FragmentFile, Iterator_ConsensusFile, Iterator_AlignmentFile, Iterator_MatrixFile, Iterator_VCFFile, Iterator_MapstatFile, Iterator_SPAFile
from cgelib.sequence.SeqHit import KMAHit
from cgelib.utils.file_mixin import CGELibFileError, ResultFile, CGELibFileParseError
from cgelib.utils.format_mixin import FormatVariable
from cgelib.sequence.Feature import AlnFeature




class KMA_Result(dict):

    FILE_ITERATOR = {
            "Result": Iterator_ResFile,
            "Fragments": Iterator_FragmentFile,
            "Consensus": Iterator_ConsensusFile,
            "Alignment": Iterator_AlignmentFile,
            "Matrix": Iterator_MatrixFile,
            "Mapstat": Iterator_MapstatFile,
            "Frag_Raw": None,
            "VCF": Iterator_VCFFile,
            "Sparse": Iterator_SPAFile
            }

    def __init__(self, filename, aln_files, output_path=None):

        if output_path is None:
            file_path = filename
        else:
            file_path = "%s/%s" % (output_path, filename)

        files = FormatVariable.is_a_list(aln_files)

        KMA_Result.check_Sparse(files)
        self.files = KMA_Result.force_file_last(files=files)

        for file in self.files:
            result_file = KMA_Result.init_file(file_path=file_path,
                                               aln_file=file)
            self[result_file] = None

    def _get_entry(self, file_type):
        for file_result in self:
            if file_result.type == file_type:
                iterator_resultfile = dict.__getitem__(self, file_result)
                try:
                    entry = next(iterator_resultfile)
                except CGELibFileParseError:
                    entry = None
                return entry
        raise KeyError("The file type %s do not exist in the KMA_result." % (
                       file_type))

    def __getitem__(self, file_type):
        for file_result in self:
            if file_result.type == file_type:
                return file_result
        raise KeyError("The file type %s do not exist in the KMA_result." % (
                       file_type))

    @staticmethod
    def check_Sparse(aln_files):
        if "Sparse" in aln_files and aln_files.__len__() > 1:
            raise CGELibFileError("File 'Sparse' is a unique result file. It "
                                  "cannot be read with other files")
    @staticmethod
    def force_file_last(files, file="VCF"):
        if "VCF" in files:
            files.sort(key='string2'.__eq__)
        return files

    @staticmethod
    def init_file(file_path, aln_file):
        if aln_file in KMA_Result.FILE_ITERATOR:
            iterator_result = KMA_Result.FILE_ITERATOR[aln_file]
        else:
            raise KeyError("The alignment file %s is not part of the KMA"
                           "alignment files." % aln_file)
        result_file = ResultFile(type=aln_file, file_path=file_path,
                                 extension=iterator_result.EXTENSION,
                                 read_method=iterator_result)
        return result_file

    def start_iterators(self):
        for key, value in self.items():
            self.__setitem__(key, key.read())

    def iterate_hits(self):
        iter_true = True
        vcf_entry = None
        # Activate iterators
        self.start_iterators()
        while iter_true:
            list_hits = []
            for kma_file in self.files:
                if kma_file is "VCF" and len(self.files) > 1:
                    if vcf_entry is None:
                        entry = self._get_entry(kma_file)
                        if entry is None:
                            return
                    else:
                        entry = vcf_entry
                    if entry["templateID"] == hit["templateID"]:
                        transl_entry = AlnFeature.translate_entry(
                                    file_iterator=self[kma_file].read_method,
                                    entry=entry)
                        file_instance = self[kma_file]
                        hit = KMAHit(orig_file=file_instance,
                                     data=transl_entry)
                        list_hits.append(hit)
                    else:
                        entry = vcf_entry
                else:
                    entry = self._get_entry(kma_file)
                    if entry is None:
                        return
                    else:
                        transl_entry = AlnFeature.translate_entry(
                                        file_iterator=self[kma_file].read_method,
                                        entry=entry)
                        file_instance = self[kma_file]
                        hit = KMAHit(orig_file=file_instance, data=transl_entry)
                        list_hits.append(hit)
            merged_hit = KMAHit.merge_hits(list_hits)
            yield merged_hit



#!/usr/bin/env python3


class CustomValueParsers():

    @staticmethod
    def char64(val):
        """
        Test val is exactely 64 characters. Can be anything that can be handled
        by str()
        """
        val = str(val)
        if(len(val) != 64):
            return ("This field expects a string of lenght 64 but the lenght "
                    "of the string is {}. The string is: {}"
                    .format(len(val), val))



# Created by Alfred Ferrer Florensa
"""Contains objects for reading """

import os
import sys
import gzip
import signal
import pandas as pd
from cgelib.utils.file_mixin import _File, ResultFile, CGELibFileParseError
from cgelib.alignment.blast.alignment_files import Iterator_XMLFile, Iterator_BlastSepFile
from cgelib.sequence.SeqHit import BlastnHit
from cgelib.sequence.Feature import AlnFeature


class BLASTN_Result(dict):

    FILE_ITERATOR = {
            "XML": Iterator_XMLFile,
            "TSV": Iterator_BlastSepFile,
            "CSV": Iterator_BlastSepFile,
            "TSV_extra": Iterator_BlastSepFile
    }

    ITERATOR_OPTIONS = {
            "XML": None,
            "TSV": {"separator": "tab",
                    "comment_lines": False},
            "CSV": {"separator": "comma",
                    "comment_lines": False},
            "TSV_extra": {"separator": "tab",
                          "comment_lines": True}
    }


    def __init__(self, filename, aln_file, output_path=None, extension=None,
                 headers=None):
        if output_path is None:
            out_prefix = ""
        else:
            out_prefix = output_path + "/"
        if extension is not None:
            file_path = out_prefix + filename + extension
        else:
            file_path = out_prefix + filename

        if not isinstance(aln_file, str):
            raise TypeError("The name of the file has to be a string")

        if not os.path.isfile(file_path):
            raise OSError("The file %s do not exists." % file_path)

        self.files = aln_file
        result_file= BLASTN_Result.init_file(file_path=file_path,
                                             aln_file=aln_file,
                                             header=headers)
        self[result_file] = None


    @staticmethod
    def init_file(file_path, aln_file, header=None):
        if aln_file in BLASTN_Result.FILE_ITERATOR:
            iterator_result = BLASTN_Result.FILE_ITERATOR[aln_file]
            options_read = BLASTN_Result.ITERATOR_OPTIONS[aln_file]
        else:
            raise KeyError("""The alignment file %s is not part of the BLASTN
                              alignment files.""" % aln_file)

        result_file = ResultFile(type=aln_file, file_path=file_path,
                                 read_method=iterator_result)
        if header is not None:
            if aln_file is "XML":
                raise ValueError("No header can be defined when using an XML"
                                 " file")
            else:
                result_file.options_read["header"] = header
        if options_read is not None:
            result_file.options_read["separator"] = options_read["separator"]
            result_file.options_read["comment_lines"] = options_read["comment_lines"]

        return result_file

    def _get_entry(self, file_type):
        for file_result in self:
            if file_result.type == file_type:
                iterator_resultfile = dict.__getitem__(self, file_result)
                try:
                    entry = next(iterator_resultfile)
                except CGELibFileParseError:
                    entry = None
                return entry
        raise KeyError("The file type %s do not exist in the BLASTN_result." % (
                       file_type))

    def start_iterators(self):
        for key, value in self.items():
            self.__setitem__(key, key.read())

    def __getitem__(self, file_type):
        for file_result in self:
            if file_result.type == file_type:
                return file_result
        raise KeyError("The file type %s do not exist in the BLASTN_result." % (
                       file_type))

    def iterate_hits(self):
        iter_true = True
        self.start_iterators()
        while iter_true:
            entry = self._get_entry(self.files)
            if entry is None:
                return 
            file_instance = self[self.files]
            try:
                empty_var = entry.pop("empty")
            except KeyError:
                empty_var = False
            transl_entry = AlnFeature.translate_entry(
                                file_iterator=file_instance.read_method,
                                entry=entry)
            hit = BlastnHit(orig_file=file_instance, empty=empty_var,
                            data=transl_entry)
            yield hit



#!/usr/bin/env python3

from collections import UserDict

from cgelib.output.valueparsers import ValueParsers


class ParserDict(UserDict):

    def __init__(self, input_parser=None):
        super().__init__(self)

        if(input_parser is None):
            parser_class = ValueParsers
        else:
            parser_class = input_parser

        val_parser_list = self.get_method_names(parser_class)

        for parser in val_parser_list:
            if(parser.startswith("parse_")):
                parse_key = parser[6:]
                self[parse_key] = getattr(parser_class, parser)
            else:
                raise SyntaxError(("A function in the {} class did "
                                   "not start with 'parse_'. Function is "
                                   "named: {}"
                                   .format(parser_class.__name__, parser)))

    @staticmethod
    def get_method_names(cls):
        return [func for func in dir(cls) if(not func.startswith("__")
                                             and callable(getattr(cls, func)))]



from cgelib.applications.command import _ContentArgument, CommandLineBase
from cgelib.applications.command import _SwitchArgument


class _BlastBaseCommandline(CommandLineBase):
    """Base Commandline object for Blast wrappers (PRIVATE).
    This is provided for subclassing, it deals with shared options
    common to all the blast tools (blastn, etc).
    """

    def __init__(self, cmd=None, path_exec="", **kwargs):
        assert cmd is not None
        extra_parameters = [
            # Core:
            _SwitchArgument(
                ["-h", "help"],
                "Print USAGE, DESCRIPTION description; "
                "ignore other arguments.",
                no_run=True,
            ),
            _SwitchArgument(
                ["-help", "help_extended"],
                "Print EXTENDED USAGE, DESCRIPTION and ARGUMENTS description; "
                "ignore other arguments.",
                no_run=True,
            ),
            _SwitchArgument(
                ["-version", "version"],
                "Print version number;  "
                "ignore other arguments.",
                no_run=True,
            ),
            _ContentArgument(
                ["", "custom_args"],
                "Add custom arguments that are not included in the "
                "kma_application.py file."
            ),
        ]
        try:
            # Insert extra parameters - at the start just in case there
            # are any arguments which must come last:
            self.parameters = extra_parameters + self.parameters
        except AttributeError:
            # Should we raise an error?  The subclass should have set this up!
            self.parameters = extra_parameters
        # Do we need a method to add a method for adding arguments in front?
        CommandLineBase.__init__(self, cmd, path_exec, **kwargs)

    def _validate_incompatibilities(self, incompatibles):
        """Validate parameters for incompatibilities (PRIVATE).
        Used by the _validate method.
        """
        for a in incompatibles:
            if self._get_parameter(a):
                for b in incompatibles[a]:
                    if self._get_parameter(b):
                        raise ValueError("Options %s and %s are incompatible."
                                         % (a, b))


class _BlastSearchBaseCommandline(_BlastBaseCommandline):
    """Base Commandline object for Blast wrappers (PRIVATE).
    This is provided for subclassing, it deals with shared options
    common to all the blast tools (blastn, etc).
    """

    def __init__(self, cmd=None, path_exec="", **kwargs):
        assert cmd is not None
        extra_parameters = [
            # Core:
            _ContentArgument(
                ["-db", "db"],
                "BLAST database name",
                filename=True,
                is_required=True,
                default=None,
                alter_options=["subject", "subject_loc"],
                incompatible=["subject", "subject_loc"],
            ),
            _ContentArgument(
                ["-query", "query"],
                "Query file name",
                filename=True,
                is_required=True,
                alter_options=["query_loc"],
                default="-",
            ),
            _ContentArgument(
                ["-query_loc", "query_location"],
                "Location on the query sequence",
                filename=True,
                is_required=True,
                alter_options=["query"],
                default="STDIN",
            ),
            _ContentArgument(
                ["-out", "output"],
                "Output file name",
                filename=True,
                is_required=True,
                default="STDOUT",
            ),
            _ContentArgument(
                ["-evalue", "evalue"],
                "Expect value (E) for saving hits",
                filename=True,
                is_required=False,
                default=10.0,
            ),
            _ContentArgument(
                ["-subject", "subject"],
                "File with subject sequence(s) to search.",
                filename=True,
                incompatible=["db", "gilist", "seqidlist", "negative_gilist",
                              "db_soft_mask", "db_hard_mask"],
                is_required=False,
                default=None,
            ),
            _ContentArgument(
                ["-subject_loc", "subject_location"],
                "Location on the subject sequence (Format: start-stop).",
                filename=True,
                incompatible=["db", "gilist", "seqidlist", "negative_gilist",
                              "db_soft_mask", "db_hard_mask"],
                is_required=False,
                default=None,
            ),
            _SwitchArgument(
                ["-show_gis", "show_GIs"],
                "Show NCBI GIs in report.",
            ),
            _ContentArgument(
                ["-num_descriptions", "num_descriptions"],
                "Show one-line descriptions for this number of database "
                "sequences.",
                filename=False,
                checker_function=lambda x: x >= 0,
                is_required=False,
                default=500,
            ),
            _ContentArgument(
                ["-num_alignments", "num_alignments"],
                "Show alignments for this number of database sequences.",
                filename=False,
                checker_function=lambda x: x >= 0,
                is_required=False,
                default=250,
            ),
            _ContentArgument(
                ["-max_target_seqs", "max_target_seqs"],
                "Number of aligned sequences to keep. Use with report formats "
                "that do not have separate definition line and alignment "
                "sections such as tabular (all outfmt > 4). Not compatible "
                "with num_descriptions or num_alignments. Ties are broken by "
                "order of sequences in the database.",
                filename=False,
                is_required=False,
                checker_function=lambda x: x >= 1,
                incompatible=["num_descriptions", "num_alignments"],
                default=500,
            ),
            _ContentArgument(
                ["-max_hsps", "max_hsps"],
                "Maximum number of HSPs (alignments) to keep for any single "
                "query-subject pair. The HSPs shown will be the best as judged"
                " by expect value. This number should be an integer that is "
                "one or greater. If this option is not set, BLAST shows all "
                "HSPs meeting the expect value criteria. Setting it to one "
                "will show only the best HSP for every query-subject pair",
                filename=False,
                checker_function=lambda x: x >= 1,
                is_required=False,
                default=None,
            ),
            _SwitchArgument(
                ["-html", "html"],
                "Produce HTML output",
            ),
            _ContentArgument(
                ["-seqidlist", "seqidlist"],
                "Restrict search of database to list of SeqId's.",
                filename=False,
                incompatible=["negative_gilist", "gilist", "remote",
                              "subject", "subject_loc"],
                is_required=False,
                default=None,
            ),
            _ContentArgument(
                ["-gilist", "gilist"],
                "Restrict search of database to GI’s listed in this file. "
                "Local searches only.",
                filename=True,
                incompatible=["negative_gilist", "seqidlist", "remote",
                              "subject", "subject_loc"],
                is_required=False,
                default=None,
            ),
            _ContentArgument(
                ["-negative_gilist", "negative_gilist"],
                "Restrict search of database to everything except the GI’s "
                "listed in this file. Local searches only.",
                filename=True,
                is_required=False,
                default=None,
            ),
            _ContentArgument(
                ["-entrez_query", "entrez_query"],
                "Restrict search with the given Entrez query. Remote searches "
                "only.",
                filename=False,
                required_options=["remote"],
                is_required=False,
                default=None,
            ),
            _ContentArgument(
                ["-culling_limit", "culling_limit"],
                "Delete a hit that is enveloped by at least this many "
                "higher-scoring hits.",
                filename=False,
                is_required=False,
                incompatible=["best_hit_overhang", "best_hit_score_edge"],
                checker_function=lambda x: x >= 0,
                default=None,
            ),
            _ContentArgument(
                ["-best_hit_overhang", "best_hit_overhang"],
                "Best Hit algorithm overhang value (recommended value: 0.1)",
                filename=False,
                incompatible=["culling_limit"],
                is_required=False,
                default=None,
            ),
            _ContentArgument(
                ["-best_hit_score_edge", "best_hit_score_edge"],
                "Best Hit algorithm score edge value (recommended value: 0.1)",
                filename=False,
                is_required=False,
                incompatible=["culling_limit"],
                default=None,
            ),
            _ContentArgument(
                ["-dbsize", "dbsize"],
                "Effective size of the database",
                filename=False,
                is_required=False,
                default=None,
            ),
            _ContentArgument(
                ["-searchsp", "searchsp"],
                "Effective length of the search space",
                filename=False,
                checker_function=lambda x: x >= 0,
                is_required=False,
                default=None,
            ),
            _ContentArgument(
                ["-import_search_strategy", "import_search_strategy"],
                "Search strategy file to read.",
                filename=True,
                incompatible=["export_search_strategy"],
                is_required=False,
                default=None,
            ),
            _ContentArgument(
                ["-export_search_strategy", "export_search_strategy"],
                "Record search strategy to this file.",
                filename=True,
                incompatible=["import_search_strategy"],
                is_required=False,
                default=None,
            ),
            _SwitchArgument(
                ["-parse_deflines", "parse_deflines"],
                "Parse query and subject bar delimited sequence identifiers "
                "(e.g., gi|129295).",
            ),
            _ContentArgument(
                ["-num_threads", "num_threads"],
                "Number of threads (CPUs) to use in blast search.",
                filename=None,
                checker_function=lambda x: x >= 1,
                incompatible=["remote"],
                is_required=False,
                default=1,
            ),
            _SwitchArgument(
                ["-remote", "remote"],
                "Execute search on NCBI servers?",
                incompatible=["gilist", "seqidlist", "negative_gilist",
                              "subject_loc", "num_threads"]
            ),
            _ContentArgument(
                ["-outfmt", "outfmt"],
                "Number of threads (CPUs) to use in blast search.",
                filename=None,
                is_required=False,
                default=0,
                allow_multiple=True,
            ),
        ]
        try:
            # Insert extra parameters - at the start just in case there
            # are any arguments which must come last:
            self.parameters = extra_parameters + self.parameters
        except TypeError:
            # Should we raise an error?  The subclass should have set this up!
            self.parameters = extra_parameters
        # Do we need a method to add a method for adding arguments in front?
        _BlastBaseCommandline.__init__(self, cmd, path_exec, **kwargs)


class BlastNCommandline(_BlastSearchBaseCommandline):
    """Base Commandline object for Blast wrappers (PRIVATE).
    This is provided for subclassing, it deals with shared options
    common to all the blast tools (blastn, etc).
    """

    def __init__(self, cmd="blastn", path_exec="", **kwargs):
        assert cmd is not None
        self.parameters = [
            _ContentArgument(
                ["-task", "task"],
                "Task to execute",
                filename=None,
                is_required=False,
                checker_function=lambda x: x in ["blastn", "blastn-short",
                                                 "dc-megablast", "megablast",
                                                 "rmblastn"],
                default="Megablast",
            ),
            _ContentArgument(
                ["-word_size", "word_size"],
                "Word size for wordfinder algorithm (length of best perfect "
                "match)",
                filename=None,
                is_required=False,
                checker_function=lambda x: x >= 4,
                default="Depends on task",
            ),
            _ContentArgument(
                ["-gapopen", "gapopen"],
                "Cost to open a gap.",
                filename=None,
                is_required=False,
                default="Depends on task",
            ),
            _ContentArgument(
                ["-gapextend", "gapextend"],
                "Word size for wordfinder algorithm (length of best perfect "
                "match)",
                filename=None,
                is_required=False,
                default="Depends on task",
            ),
            _ContentArgument(
                ["-reward", "reward"],
                "Reward for a nucleotide match.",
                filename=None,
                is_required=False,
                checker_function=lambda x: x >= 0,
                default="Depends on task",
            ),
            _ContentArgument(
                ["-penalty", "penalty"],
                "Penalty for a nucleotide mismatch.",
                filename=None,
                is_required=False,
                checker_function=lambda x: x <= 0,
                default="Depends on task",
            ),
            _ContentArgument(
                ["-strand", "strand"],
                "Query strand(s) to search against database/subject. Choice of"
                "both, minus, or plus.",
                filename=None,
                is_required=False,
                checker_function=lambda x: x in ["both", "minus", "plus"],
                default="both",
            ),
            _ContentArgument(
                ["-dust", "dust"],
                "Filter query sequence with dust.",
                filename=None,
                is_required=False,
                default="20 64 1",
            ),
            _ContentArgument(
                ["-filtering_db", "filtering_db"],
                "Mask query using the sequences in this database.",
                filename=None,
                is_required=False,
                default=None,
            ),
            _ContentArgument(
                ["-window_masker_taxid", "window_masker_taxid"],
                "Enable WindowMasker filtering using a Taxonomic ID.",
                filename=None,
                is_required=False,
                default=None,
            ),
            _ContentArgument(
                ["-sum_stats", "sum_stats"],
                "Use sum statistics.",
                filename=None,
                is_required=False,
                default=None,
            ),
            _ContentArgument(
                ["-window_masker_db", "window_masker_db"],
                "Enable WindowMasker filtering using this file.",
                filename=None,
                is_required=False,
                default=None,
            ),
            _SwitchArgument(
                ["-soft_masking", "soft_masking"],
                "Apply filtering locations as soft masks (i.e., only for "
                "finding initial matches).",
            ),
            _SwitchArgument(
                ["-lcase_masking", "lcase_masking"],
                "Use lower case filtering in query and subject sequence(s).",
            ),
            _ContentArgument(
                ["-db_soft_mask", "db_soft_mask"],
                "Filtering algorithm ID to apply to the BLAST database as soft"
                "mask (i.e., only for finding initial matches).",
                filename=None,
                incompatible=["db_hard_mask", "subject", "subject_loc"],
                is_required=False,
                default=True,
            ),
            _ContentArgument(
                ["-db_hard_mask", "db_hard_mask"],
                "Filtering algorithm ID to apply to the BLAST database as "
                "hard mask (i.e., sequence is masked for all phases of search)"
                ".",
                filename=None,
                incompatible=["db_hard_mask", "subject", "subject_loc"],
                is_required=False,
                default=None,
            ),
            _ContentArgument(
                ["-perc_identity", "perc_identity"],
                "Percent identity cutoff.",
                filename=None,
                is_required=False,
                default=0,
            ),
            _ContentArgument(
                ["-qcov_hsp_perc", "qcov_hsp_perc"],
                "Percent query coverage per hsp.",
                filename=None,
                is_required=False,
                default=0,
            ),
            _ContentArgument(
                ["-template_type", "template_type"],
                "Discontiguous MegaBLAST template type. Allowed values are "
                "coding, optimal and coding_and_optimal.",
                filename=None,
                is_required=False,
                checker_function=lambda x: x in ["coding",
                                                 "coding_and_optimal",
                                                 "optimal"],
                required_options=["template_length"],
                default=0,
            ),
            _ContentArgument(
                ["-template_length", "template_length"],
                "Discontiguous MegaBLAST template length.",
                filename=None,
                required_options=["template_length"],
                is_required=False,
                default=18,
            ),
            _ContentArgument(
                ["-use_index", "use_index"],
                "Use MegaBLAST database index. Indices may be created with the"
                " makembindex application.",
                filename=None,
                is_required=False,
                default=18,
            ),
            _ContentArgument(
                ["-index_name", "index_name"],
                "MegaBLAST database index name.",
                filename=None,
                is_required=False,
                default=18,
            ),
            _ContentArgument(
                ["-xdrop_ungap", "xdrop_ungap"],
                "Heuristic value (in bits) for ungapped extensions.",
                filename=None,
                is_required=False,
                default=20,
            ),
            _ContentArgument(
                ["-xdrop_gap", "xdrop_gap"],
                "Heuristic value (in bits) for preliminary gapped extensions.",
                filename=None,
                is_required=False,
                default=30,
            ),
            _ContentArgument(
                ["-xdrop_gap_final", "xdrop_gap_final"],
                "Heuristic value (in bits) for final gapped alignment.",
                filename=None,
                is_required=False,
                default=100,
            ),
            _SwitchArgument(
                ["-no_greedy", "no_greedy"],
                "Use non-greedy dynamic programming extension.",
            ),
            _ContentArgument(
                ["-min_raw_gapped_score", "min_raw_gapped_score"],
                "Minimum raw gapped score to keep an alignment in the "
                "preliminary gapped and trace-back stages. Normally set based "
                "upon expect value.",
                filename=None,
                is_required=False,
                default=30,
            ),
            _SwitchArgument(
                ["-ungapped", "ungapped"],
                "Perform ungapped alignment.",
            ),
            _ContentArgument(
                ["-window_size", "window_size"],
                "Multiple hits window size, use 0 to specify 1-hit algorithm.",
                filename=None,
                is_required=False,
                default=40,
                checker_function=lambda x: x >= 0,
            ),
            _ContentArgument(
                ["-off_diagonal_range", "off_diagonal_range"],
                "Number of off-diagonals to search for the 2nd hit, use 0 to "
                "turn off",
                filename=None,
                is_required=False,
                default=40,
                checker_function=lambda x: x >= 0,
            ),
            _ContentArgument(
                ["-line_length", "line_length"],
                "Line length for formatting alignments. Not applicable for "
                "outfmt > 4",
                filename=None,
                is_required=False,
                default=60,
                checker_function=lambda x: x >= 1,
            ),
        ]
        _BlastSearchBaseCommandline.__init__(self, cmd, path_exec, **kwargs)



from cgelib.applications.command import _ContentArgument, CommandLineBase
from cgelib.applications.command import _SwitchArgument, _SwitchValueArgument


class _KmaBaseCommandline(CommandLineBase):
    """Base Commandline object for KMA wrappers (PRIVATE).
    This is provided for subclassing, it deals with shared options
    common to all the KMA tools (kma, kma_index, kma_update, etc
    AND kma_shm).
    """

    def __init__(self, cmd=None, path_exec="", **kwargs):
        assert cmd is not None
        extra_parameters = [
            # Core:
            _SwitchArgument(
                ["-h", "h"],
                "Print USAGE, DESCRIPTION and ARGUMENTS description; "
                "ignore other arguments.",
                no_run=True,
            ),
            _SwitchArgument(
                ["-v", "v"],
                "Print version number;  "
                "ignore other arguments.",
                no_run=True,
            ),
            _ContentArgument(
                ["", "custom_args"],
                "Add custom arguments that are not included in the "
                "kma_application.py file."
            ),
        ]
        try:
            # Insert extra parameters - at the start just in case there
            # are any arguments which must come last:
            self.parameters = extra_parameters + self.parameters
        except TypeError:
            # Should we raise an error?  The subclass should have set this up!
            self.parameters = extra_parameters
        # Do we need a method to add a method for adding arguments in front?
        CommandLineBase.__init__(self, cmd, path_exec, **kwargs)

    def _validate_incompatibilities(self, incompatibles):
        """Validate parameters for incompatibilities (PRIVATE).
        Used by the _validate method.
        """
        for a in incompatibles:
            if self._get_parameter(a):
                for b in incompatibles[a]:
                    if self._get_parameter(b):
                        raise ValueError("Options %s and %s are incompatible."
                                         % (a, b))


class KMACommandline(_KmaBaseCommandline):
    """Base Commandline object for the wrapper of the KMA aligner.
    This is provided for subclassing, it deals with shared options
    common to all the BLAST tools (blastn, rpsblast, rpsblast, etc).
    """

    def __init__(self, cmd="kma", path_exec="", **kwargs):
        assert cmd is not None
        self.parameters = [
            # Input query options:
            _ContentArgument(
                ["-i", "input"],
                "Input file name(s)",
                filename=True,
                is_required=True,
                default="STDIN",
                alter_options=["input_int", "input_ipe"],
                incompatible=["input_int", "input_ipe"],
                allow_multiple=True,
            ),
            _ContentArgument(
                ["-ipe", "input_ipe"],
                "Input paired end file name(s)",
                filename=True,
                is_required=True,
                alter_options=["input", "input_int"],
                incompatible=["input", "input_int"],
                allow_multiple=True,
            ),
            _ContentArgument(
                ["-int", "input_int"],
                "Input interleaved file name(s)",
                filename=True,
                is_required=True,
                alter_options=["input", "input_ipe"],
                incompatible=["input", "input_ipe"],
                allow_multiple=True,
            ),  # Should this be required?
            _ContentArgument(
                ["-o", "output"],
                "Output file",
                filename=True,
                is_required=True,
                default="None",
            ),
            _ContentArgument(
                ["-t_db", "template_db"],
                "Template DB",
                filename=True,
                default="None",
                is_required=True,
            ),
            _ContentArgument(
                ["-k", "k_size"],
                "Kmersize (default by db)",
                filename=False,
                default="DB defined",
                is_required=False,
            ),
            _ContentArgument(
                ["-ml", "min_len"],
                "Minimum alignment length",
                filename=False,
                default=16,
                is_required=False,
            ),
            _ContentArgument(
                ["-p", "p_value"],
                "p-value",
                filename=False,
                default=0.05,
                is_required=False,
            ),
            _ContentArgument(
                ["-ConClave", "con_clave"],
                "ConClave version",
                filename=False,
                default=1,
                is_required=False,
            ),
            _SwitchArgument(
                ["-mem_mode", "mem_mode"],
                "Use kmers to choose best template, and save memory",
            ),
            _SwitchValueArgument(
                ["-proxi", "proxi"],
                "Use proximity scoring under template mapping",
                filename=False,
                default="False/1.0",
                is_required=False,
            ),
            _SwitchArgument(
                ["-ex_mode", "ex_mode"],
                "Searh kmers exhaustively",
            ),
            _SwitchArgument(
                ["-ef", "extra_files"],
                "Output extra files",
            ),
            _SwitchValueArgument(
                ["-vcf", "vcf"],
                "Make vcf file, 2 to apply FT",
                filename=False,
                default="False/0",
                is_required=False,
            ),
            _SwitchValueArgument(
                ["-sam", "sam"],
                "Output sam to stdout, 4 to output mapped reads,"
                "2096 for aligned",
                filename=False,
                default="False/1.0",
                is_required=False,
            ),
            _SwitchArgument(
                ["-nc", "non_consensus"],
                "No consensus file",
            ),
            _SwitchArgument(
                ["-na", "no_aln"],
                "No aln file",
            ),
            _SwitchArgument(
                ["-nf", "no_frag"],
                "No frag file",
            ),
            _SwitchArgument(
                ["-deCon", "deCon"],
                "Remove contamination",
            ),
            _SwitchArgument(
                ["-dense", "dense"],
                "Do not allow insertions in assembly",
            ),
            _SwitchArgument(
                ["-sasm", "sasm"],
                "Skip alignment and assembly",
            ),
            _SwitchValueArgument(
                ["-ref_fsa", "ref_sa"],
                "Consensus sequence will have 'n' instead of gaps",
                filename=False,
                is_required=False,
                default="False/0",
            ),
            _SwitchArgument(
                ["-matrix", "matrix"],
                "Outputs assembly matrix",
            ),
            _SwitchArgument(
                ["-a", "best_maps"],
                "Print all best mappings",
            ),
            _ContentArgument(
                ["-mp", "min_phred"],
                "Minimum phred score",
                filename=False,
                default=20,
                is_required=False,
            ),
            _ContentArgument(
                ["-5p", "cut_5p"],
                "Cut a constant number of nucleotides from the 5 prime",
                filename=False,
                default=0,
                is_required=False,
            ),
            _ContentArgument(
                ["-3p", "cut_3p"],
                "Cut a constant number of nucleotides from the 3 prime",
                filename=False,
                default=0,
                is_required=False,
            ),
            _SwitchArgument(
                ["-Sparse", "sparse"],
                "Only count kmers",
            ),
            _SwitchValueArgument(
                ["-Mt1", "Mt1"],
                "Map only to 'num' template",
                filename=False,
                is_required=False,
                default="False/0",
            ),
            _ContentArgument(
                ["-ID", "ID"],
                "Minimum ID",
                filename=False,
                default="1.0%",
                is_required=False,
            ),
            _ContentArgument(
                ["-ss", "ss"],
                "Sparse sorting (q,c,d)",
                filename=False,
                default="q",
                checker_function=lambda x: x in ["q", "c", "d"],
                is_required=False,
            ),
            _ContentArgument(
                ["-pm", "pm"],
                "Pairing method (p,u,f)",
                filename=False,
                default="u",
                checker_function=lambda x: x in ["p", "u", "f"],
                is_required=False,
            ),
            _ContentArgument(
                ["-fpm", "fpm"],
                "Fine Pairing method (p,u,f)",
                filename=False,
                checker_function=lambda x: x in ["p", "u", "f"],
                default="u",
                is_required=False,
            ),
            _ContentArgument(
                ["-apm", "apm"],
                "Sets both pm and fpm (p,u,f)",
                filename=False,
                default="u",
                checker_function=lambda x: x in ["p", "u", "f"],
                is_required=False,
            ),
            _ContentArgument(
                ["-shm", "shm"],
                "Use shared DB made by kma_shm",
                filename=False,
                default="0 (lvl)",
                is_required=False,
            ),
            _SwitchArgument(
                ["-mmap", "mmap"],
                "Memory map *.comp.by",
            ),
            _ContentArgument(
                ["-tmp", "tmp"],
                "Set directory for temporary files.",
                filename=True,
                is_required=False,
            ),
            _SwitchArgument(
                ["-1t1", "kma_1t1"],
                "Force end to end mapping"
            ),
            _SwitchArgument(
                ["-hmm", "hmm"],
                "Use a HMM to assign template(s) to query sequences"
            ),
            _SwitchArgument(
                ["-ck", "count_k"],
                "Count kmers instead of pseudo alignment"
            ),
            _SwitchArgument(
                ["-ca", "circular_aln"],
                "Make circular alignments"
            ),
            _SwitchArgument(
                ["-boot", "bootstrap"],
                "Bootstrap sequence"
            ),
            _SwitchArgument(
                ["-bc", "bc"],
                "Base calls should be significantly overrepresented",
                default=True,
                incompatible=["bc90"],
            ),
            _SwitchArgument(
                ["-bc90", "bc90"],
                "Base calls should be both significantly overrepresented, and"
                " have 90% agreement.",
                incompatible=["bc"]
            ),
            _SwitchArgument(
                ["-bcNano", "bcNano"],
                "Call bases at suspicious deletions, made for nanopore."
            ),
            _ContentArgument(
                ["-bcd", "bcd"],
                "Minimum depth at base.",
                filename=False,
                default=1,
                is_required=False,
            ),
            _SwitchArgument(
                ["-bcg", "bcg"],
                "Maintain insignificant gaps",
            ),
            _SwitchArgument(
                ["-and", "and"],
                "Both mrs and p_value thresholds has to reached to in order "
                "to	report a template hit.",
            ),
            _ContentArgument(
                ["-mq", "MinMapQ"],
                "Minimum mapping quality",
                filename=False,
                default=0,
                is_required=False,
            ),
            _ContentArgument(
                ["-mrs", "MinAlnQ"],
                "Minimum alignment score, normalized to alignment length",
                filename=False,
                default=0.50,
                is_required=False,
            ),
            _ContentArgument(
                ["-mct", "mct"],
                "Max overlap between templates",
                filename=False,
                default=0.50,
                is_required=False,
            ),
            _ContentArgument(
                ["-reward", "reward"],
                "Score for match",
                filename=False,
                default=1,
                is_required=False,
            ),
            _ContentArgument(
                ["-penalty", "penalty"],
                "Penalty for mismatch",
                filename=False,
                default=-2,
                is_required=False,
            ),
            _ContentArgument(
                ["-gapopen", "gap_open"],
                "Penalty for gap opening",
                filename=False,
                default=-3,
                is_required=False,
            ),
            _ContentArgument(
                ["-gapextend", "gap_extend"],
                "Penalty for gap extension",
                filename=False,
                default=-1,
                is_required=False,
            ),
            _ContentArgument(
                ["-per", "reward_pairing"],
                "Reward for pairing end",
                filename=False,
                default=7,
                is_required=False,
            ),
            _ContentArgument(
                ["-localopen", "local_open"],
                "Penalty for openning a local chain",
                filename=False,
                default=-6,
                is_required=False,
            ),
            _ContentArgument(
                ["-Npenalty", "n_penalty"],
                "Penalty matching N",
                filename=False,
                default=0,
                is_required=False,
            ),
            _ContentArgument(
                ["-transition", "transition"],
                "Penalty for transition",
                filename=False,
                default=-2,
                is_required=False,
            ),
            _ContentArgument(
                ["-transversion", "transversion"],
                "Penalty for transversion",
                filename=False,
                default=-2,
                is_required=False,
            ),
            _SwitchArgument(
                ["-cge", "cge"],
                "Set CGE penalties and rewards",
            ),
            _ContentArgument(
                ["-t", "threads"],
                "Number of threads",
                filename=False,
                default=1,
                is_required=False,
            ),
            _SwitchArgument(
                ["-status", "status"],
                "Extra status",
            ),
            _SwitchArgument(
                ["-verbose", "verbose"],
                "Extra verbose",
            ),
            _SwitchArgument(
                ["-c", "citation"],
                "Citation",
                no_run=True,
            ),
            _SwitchArgument(
                ["-ont", "ont"],
                "Set 3rd gen genefinding preset",
            ),
            _ContentArgument(
                ["-md", "md"],
                "minimum depth",
                filename=False,
                default=0.0,
                is_required=False,
            ),
        ]
        _KmaBaseCommandline.__init__(self, cmd, path_exec, **kwargs)


class KMAIndexCommandline(_KmaBaseCommandline):
    """Base Commandline object for (new) NCBI BLAST+ wrappers (PRIVATE).
    This is provided for subclassing, it deals with shared options
    common to all the BLAST tools (blastn, rpsblast, rpsblast, etc).
    """

    def __init__(self, cmd="kma_index", path_exec="", **kwargs):
        assert cmd is not None
        self.parameters = [
            _ContentArgument(
                ["-i", "input"],
                "Input/query file name (STDIN: '--')",
                filename=True,
                default="None",
                is_required=False,
                allow_multiple=True,
                alter_options=["batch"],
                incompatible=["batch"],

            ),
            _ContentArgument(
                ["-o", "output"],
                "Output file",
                filename=False,
                default="Input file",
                is_required=False,
            ),
            _ContentArgument(
                ["-batch", "batch"],
                "Batch input file",
                filename=True,
                is_required=False,
                alter_options=["input"],
                incompatible=["input"]
            ),
            _SwitchValueArgument(
                ["-deCon", "deCon"],
                "File with contamination (STDIN: '--')",
                default="None/False",
                filename=True,
                is_required=False,
            ),
            _ContentArgument(
                ["-batchD", "batchD"],
                "Batch decon file",
                filename=True,
                is_required=False,
            ),
            _SwitchValueArgument(
                ["-t_db", "template_db"],
                "Add to existing DB",
                filename=True,
                default="None/False",
                is_required=False,
            ),
            _ContentArgument(
                ["-k", "k_size"],
                "Kmersize",
                filename=False,
                default=16,
                is_required=False,
            ),
            _ContentArgument(
                ["-k_t", "k_temp"],
                "Kmersize for template identification",
                filename=False,
                default=16,
                is_required=False,
            ),
            _ContentArgument(
                ["-k_i", "k_index"],
                "Kmersize for indexing",
                filename=False,
                default=16,
                is_required=False,
            ),
            _ContentArgument(
                ["-ML", "min_len"],
                "Minimum length for templates",
                filename=False,
                default="kmersize (16)",
                is_required=False,
            ),
            _ContentArgument(
                ["-CS", "cs"],
                "Start chain size",
                filename=False,
                default="1 M",
                is_required=False,
            ),
            _SwitchArgument(
                ["-ME", "mega_db"],
                "Mega DB",
            ),
            _SwitchArgument(
                ["-NI", "no_index"],
                "Do not dump *.index.b",
            ),
            _SwitchValueArgument(
                ["-Sparse", "sparse"],
                "Make Sparse DB ('-' for no prefix)",
                filename=False,
                default="None/False",
                is_required=False,
            ),
            _ContentArgument(
                ["-ht", "homology_temp"],
                "Homology template",
                filename=False,
                default=1.0,
                is_required=False,
            ),
            _ContentArgument(
                ["-hq", "homology_query"],
                "Homology query",
                filename=False,
                default=1.0,
                is_required=False,
            ),
            _SwitchArgument(
                ["-and", "and"],
                "Both homolgy thresholds has to be reached",
            ),
            _SwitchArgument(
                ["-nbp", "no_bias"],
                "No bias print",
            ),
        ]
        _KmaBaseCommandline.__init__(self, cmd, path_exec, **kwargs)


class KMAShmCommandline(_KmaBaseCommandline):
    """Base Commandline object for (new) NCBI BLAST+ wrappers (PRIVATE).
    This is provided for subclassing, it deals with shared options
    common to all the BLAST tools (blastn, rpsblast, rpsblast, etc).
    """

    def __init__(self, cmd="kma_shm", path_exec="", **kwargs):
        assert cmd is not None
        self.parameters = [
            _ContentArgument(
                ["-t_db", "template_db"],
                "Template DB",
                filename=True,
                default=None,
                is_required=True,
            ),
            _SwitchArgument(
                ["-destroy", "destroy"],
                "Destroy shared memory",
            ),
            _ContentArgument(
                ["-shmLvl", "shmLvl"],
                "Level of shared memory",
                filename=False,
                default=1,
                is_required=False,
            ),
            _SwitchArgument(
                ["-shm-h", "shm_h"],
                "Explain of shared memory",
            ),
        ]
        _KmaBaseCommandline.__init__(self, cmd, path_exec, **kwargs)


class KMASeq2FastaCommandline(_KmaBaseCommandline):
    """Base Commandline object for (new) NCBI BLAST+ wrappers (PRIVATE).
    This is provided for subclassing, it deals with shared options
    common to all the BLAST tools (blastn, rpsblast, rpsblast, etc).
    """

    def __init__(self, cmd="kma seq2fasta", path_exec="", **kwargs):
        assert cmd is not None
        self.parameters = [
            _ContentArgument(
                ["-t_db", "template_db"],
                "Template DB",
                filename=True,
                default=None,
                is_required=True,
            ),
            _SwitchArgument(
                ["-seqs", "seqs"],
                "Comma separated list of templates",
            ),
        ]
        _KmaBaseCommandline.__init__(self, cmd, path_exec, **kwargs)


class KMADistCommandline(_KmaBaseCommandline):
    """Base Commandline object for (new) NCBI BLAST+ wrappers (PRIVATE).
    This is provided for subclassing, it deals with shared options
    common to all the BLAST tools (blastn, rpsblast, rpsblast, etc).
    """

    def __init__(self, cmd="kma dist", path_exec="", **kwargs):
        assert cmd is not None
        self.parameters = [
            _ContentArgument(
                ["-t_db", "template_db"],
                "Template DB",
                filename=True,
                is_required=True,
            ),
            _ContentArgument(
                ["-o", "output"],
                "Output file",
                filename=True,
                default="DB",
                is_required=False,
            ),
            _ContentArgument(
                ["-f", "output_f"],
                "Output flags",
                filename=False,
                default=1,
                checker_function=lambda x: x in [1, 4],
                is_required=False,
            ),
            _SwitchArgument(
                ["-fh", "help_flags"],
                "Help on option '-f'",
            ),
            _ContentArgument(
                ["-d", "distance_method"],
                "DistanceMethod",
                filename=False,
                checker_function=lambda x: x in [1, 2, 4, 8, 16, 32, 64, 128,
                                                 256, 512, 1024, 2048, 4096],
                default=1,
                is_required=False,
            ),
            _SwitchArgument(
                ["-dh", "help_distance"],
                "Help on option '-d'",
            ),
            _SwitchArgument(
                ["-m", "matrix"],
                "Allocate matrix on the disk",
            ),
            _ContentArgument(
                ["-tmp", "tmp"],
                "Set directory for temporary file",
                filename=True,
                is_required=False,
            ),
            _ContentArgument(
                ["-t", "threads"],
                "Number of threads",
                filename=False,
                default=1,
                is_required=False,
            ),
        ]
        _KmaBaseCommandline.__init__(self, cmd, path_exec, **kwargs)


class KMADBCommandline(_KmaBaseCommandline):
    """Base Commandline object for (new) NCBI BLAST+ wrappers (PRIVATE).
    This is provided for subclassing, it deals with shared options
    common to all the BLAST tools (blastn, rpsblast, rpsblast, etc).
    """

    def __init__(self, cmd="kma db", path_exec="", **kwargs):
        assert cmd is not None
        self.parameters = [
            _ContentArgument(
                ["-t_db", "template_db"],
                "Template DB",
                filename=True,
                is_required=True,
            ),
        ]
        _KmaBaseCommandline.__init__(self, cmd, path_exec, **kwargs)



# Created by Alfred Ferrer Florensa
"""Contains objects for reading """

#from cgelib.alignment.blast.read_files import BLASTN_Result
from cgelib.alignment.KMA.read_files import KMA_Result
from cgelib.alignment.blast.read_files import BLASTN_Result
from cgelib.utils.savers_mixin import SaveJson
from cgelib.utils.file_mixin import CGELibFileParseError
from cgelib.output.result import Result
from cgelib.sequence.SeqHit import AlnHit, CGEHitValueError


class _Alignment:

    def __init__(self, filenames, result_file, aligner=None, output_path=None,
                 command=None, stdout=None, stderr=None):
        self.output_path = output_path
        self.filenames = filenames
        self.file = result_file
        self.aligner = aligner
        self.init_result = Result.init_software_result(name="cgelib",
                                                       gitdir=None,
                                                       type="python_module")


    def get_alignment_results(self):

        self._load_alignment_results()

        return self.init_result

    def _load_alignment_results(self):
        try:
            iterator_results = self.parse_hits()
        except AttributeError:
            iterator_results = []

        for hit in iterator_results:
            hit._get_unique_hit_key(hit_collection=self.init_result["aln_hits"])
            result_hit = AlnHit.dfs_to_dict(hit)
            self.init_result.add_class(cl="aln_hits", **result_hit)

    def save_alignment(self, json_path):

        SaveJson.dump_json(std_result_file=json_path,
                            std_result=self.init_result)


class KMAAlignment(_Alignment):

    def __init__(self, filenames, result_file, output_path=None):

        _Alignment.__init__(self, output_path=output_path, filenames=filenames,
                            result_file=result_file, aligner="kma")

    def parse_hits(self):
        for file in self.filenames:
            KMA_aln = KMA_Result(output_path=self.output_path, filename=file,
                                 aln_files=self.file)
            iter_aln = KMA_aln.iterate_hits()
            file_open = True
            while file_open:
                hit = None
                try:
                    hit = next(iter_aln)
                except StopIteration:
                    file_open = False
                if hit is not None:
                    yield hit


class BlastNAlignment(_Alignment):

    def __init__(self, output_path, filenames, result_file, extension=None,
                 headers=None):

        self.extension = extension
        self.headers = headers
        if not isinstance(result_file, str):
            raise CGEHitValueError("Only one result file type from an "
                                   "alignment can be read")

        _Alignment.__init__(self, output_path=output_path, filenames=filenames,
                            result_file=result_file, aligner="blastn")

    def parse_hits(self):
        for file in self.filenames:
            BLASTN_aln = BLASTN_Result(output_path=self.output_path,
                                       filename=file, aln_file=self.file,
                                       extension=self.extension,
                                       headers=self.headers)
            iter_aln = BLASTN_aln.iterate_hits()
            file_open = True
            while file_open:
                hit = None
                try:
                    hit = next(iter_aln)
                except StopIteration:
                    file_open = False
                if hit is not None:
                    yield hit



#!/usr/bin/env python3

from collections import UserDict


class TransformerMixin():

    @staticmethod
    def userdict_to_dict(userdict):
        """
            Input: dict or UserDict object.
            Output: same as input, but all UserDict objects (also the ones
                    stored inside other UserDict or dict objects) have been
                    converted to dict objects.
        """
        builtin_dict = dict()
        for key, val in userdict.items():
            if(isinstance(val, UserDict) or isinstance(val, dict)):
                builtin_dict[key] = TransformerMixin.userdict_to_dict(val)
            else:
                builtin_dict[key] = val
        return builtin_dict



# Created by Alfred Ferrer Florensa
"""Contains objects for reading """

import os
import glob
import json
from cgelib.utils.savers_mixin import SaveJson
from cgelib.applications.KMA import KMACommandline
from cgelib.applications.Blast import BlastNCommandline
from cgelib.alignment.read_alignment import KMAAlignment, BlastNAlignment
from cgelib.output.result import Result
from cgelib.sequence.SeqHit import AlnHit
from cgelib.applications.command import CommandResume
from cgelib.alignment.blast.alignment_files import Iterator_BlastSepFile



class CGELibAttributeError(AttributeError):
    """"""

    def __init__(self, message, *args):
        self.message = message
        # allow users initialize misc. arguments as any other builtin Error
        super(CGELibAttributeError, self).__init__(message, *args)



class _Aligner:

    def __init__(self, aligner=None, commandline_exec=None, result_file=None):

        self.aligner = aligner

        self.commandline_exec = commandline_exec
        self.result_file = result_file
        self.alignment = None
        self.alignment_parameters = None
        self.result_dict = Result.init_software_result(name="cgelib",
                                                       gitdir=None,
                                                       type="python_module")
        self.alignment_files = []

    def load_exec_command(self, command, stdout, stderr, key=None):

        command_info = CommandResume(exec_name=self.aligner, command=command,
                                     stdout=stdout, stderr=stderr)

        CommandResume.std_executable(cmdresume=command_info,
                                     json_report=self.result_dict, key=key)

        self.result_dict.add_class(cl="software_executions", **command_info)

    def create_commands(self, parameters, output_param, variable_iter=None, values_iter=None):
        commandlines = {}
        if self.commandline_exec is None:
            raise ValueError("No commandline executable has been defined")
        if variable_iter is None:
            commandlines[variable_iter] = self.commandline_exec(**parameters)
        else:
            if variable_iter in parameters:
                raise KeyError("The parameter to iterate over the results (%s)"
                               " has already been set with 'set_aligner_params"
                               "' creating the commandline (%s)" % (
                                        variable_iter, parameters))
                #variable_list = parameters[variable_iter]
            else:
                variable_list = values_iter
                #raise KeyError("The parameter to iterate over the results (%s)"
                #               " is not found in the parameters introduced for"
                #               " creating the commandline (%s)" % (variable_iter,
                #                                                   parameters))
            if not isinstance(variable_list, list):
                raise ValueError("The variable chosen (%s) to iterate over is "
                                 " not a list (%s)" % (variable_iter, variable_list)
                                 )
            else:
                init_out_path = parameters[output_param]
                for variable_value in variable_list:
                    parameters[variable_iter] = variable_value
                    output_path = _Aligner.create_outputname(
                                    aligner=self.aligner,
                                    out_path=init_out_path,
                                    parameters=parameters,
                                    variable=variable_iter)
                    key_path = os.path.basename(output_path).split(os.extsep, 1)[0]
                    parameters[output_param] = output_path
                    self.alignment_files.append(output_path)
                    commandlines[key_path] = self.commandline_exec(**parameters)
        return commandlines

    @staticmethod
    def create_outputname(aligner, out_path, parameters, variable):
        value = parameters[variable]
        if glob.glob(value+".*"):
            for file in glob.glob(value+".*"):
                value = os.path.basename(file).split(os.extsep, 1)[0]
                break
        elif os.path.isfile(value):
            value = os.path.splitext(os.path.basename(value))[0]
        else:
            value = value
        if os.path.isdir(out_path):
            path_str = "%s/%s-alignment_" % (out_path, aligner)
        else:
            path_str = "%s_%s-alignment_" % (out_path, aligner)
        if aligner == "kma":
            new_out = "%s%s-%s" % (path_str, variable, value)
        else:
            extension = BlastNAligner._create_extension(parameters)
            new_out = "%s%s-%s%s" % (path_str, variable, value, extension)
        return os.path.abspath(new_out)

    def report_aligner(self):
        if not self.result_dict["software_exec"]:
            raise AttributeError("""The Report of the aligner is empty.
                                    Probably as the aligner has not been runned
                                    """)
        return self.result_dict["software_exec"]

    @staticmethod
    def define_alignment_params(**kwargs):
        alignment_parameters = {}
        alignment_parameters.update((key,
                                     value) for key, value in kwargs.items())
        return alignment_parameters

    def set_aligner_params(self, **kwargs):

        self.alignment_parameters = _Aligner.define_alignment_params(**kwargs)


    def __call__(self, commands):

        for key, run_command in commands.items():
            stdout, stderr = run_command()
            self.load_exec_command(command=str(run_command), stdout=stdout,
                                   stderr=stderr, key=key)


    def parse_hits(self):

        if self.alignment is None:
            raise CGELibAttributeError("Alignment has not been created yet. "
                                       "Run the Aligner to produce the "
                                       "alignment")

        iterator_alignment = self.alignment.parse_hits()
        for hit in iterator_alignment:
            yield hit

    def _load_alignment_results(self):
        try:
            iterator_results = self.parse_hits()
        except AttributeError:
            iterator_results = []

        for hit in iterator_results:
            hit._get_unique_hit_key(hit_collection=self.result_dict["aln_hits"])
            result_hit = AlnHit.dfs_to_dict(hit)
            hit["exec_key"] = AlnHit.get_run_key(
                        software_execs=self.result_dict["software_executions"],
                        hit=hit
                        )
            self.result_dict.add_class(cl="aln_hits", **result_hit)

    def read_alignment_result(self):

        self._load_alignment_results()

        return self.result_dict

    def save_results(self, json_path, std_result=None):
        if std_result is None:
            SaveJson.dump_json(std_result_file=json_path,
                                std_result=self.result_dict)
        else:
            SaveJson.dump_json(std_result_file=json_path,
                                std_result=std_result)

    @classmethod
    def load_results(cls, json_path):

        if not os.path.isfile(json_path):
            raise OSError("The json file %s does not exists" % json_path)
        with open(json_path, "r") as read_file:
            data = json.load(read_file)

        if cls.__name__ == "KMAAligner":
            new_aligner_class = KMAAligner()
        elif cls.__name__ == "BlastNAligner":
            new_aligner_class = BlastNAligner()
        else:
            new_aligner_class = _Aligner()
        for key, value in data["software_executions"].items():
            new_aligner_class.result_dict.add_class(cl="software_executions",
                                                **value)
        print(data["aln_hits"])
        new_aligner_class.alignment = data["aln_hits"]

        return new_aligner_class




class BlastNAligner(_Aligner):

    outfmt_to_resultfile = {
        5: "XML", 6: "TSV", 7: "TSV_extra", 10: "CSV"
    }

    std_header = ["qseqid", "sseqid", "pident", "length", "mismatch",
                    "gapopen", "qstart", "qend", "sstart", "send",
                    "evalue", "bitscore"]

    def __init__(self, result_file=None):

        _Aligner.__init__(self, aligner="blastn",
                          commandline_exec=BlastNCommandline,
                          result_file=result_file)

    @staticmethod
    def get_output_format(alignment_parameters):
        if "outfmt" in alignment_parameters:
            if isinstance(alignment_parameters["outfmt"], list):
                try:
                    file_format = int(alignment_parameters["outfmt"][0])
                    header = alignment_parameters["outfmt"][1:]
                except ValueError:
                    raise ValueError("The list of 'outfmt' has to have the number"
                                     " of the alignment view option first "
                                     "(%s)" % (
                                     alignment_parameters["outfmt"][0]))
            else:
                file_format = int(alignment_parameters["outfmt"])
                header = None
        else:
            raise ValueError("The package CGELIB cannot read the "
                             "results of running Blast with the parameter "
                             "'-outfmt' set to 0 (standard).")
        return file_format, header

    @staticmethod
    def get_header_format(alignment_parameters):
        file_format, header_format = BlastNAligner.get_output_format(
                                    alignment_parameters=alignment_parameters)
        if file_format is 5:
            header = None
        else:
            if header_format is None:
                header = BlastNAligner.std_header
            else:
                header = []
                for i in header_format:
                    if i == "std":
                        header.extend(BlastNAligner.std_header)
                    else:
                        if i not in Iterator_BlastSepFile.STD_header.keys():
                            raise ValueError("The format specifier %s is not"
                                             " part of the blastn options" % (
                                             i))
                        else:
                            if i not in header:
                                header.append(i)
        return header

    def get_resultfiles(self):

        file_format, _ = BlastNAligner.get_output_format(
                                alignment_parameters=self.alignment_parameters)

        if self.result_file is None:
            if(file_format in
               BlastNAligner.outfmt_to_resultfile):
                self.result_file = BlastNAligner.outfmt_to_resultfile[
                                        self.alignment_parameters["outfmt"]]
            else:
                raise ValueError("""The package CGELIB cannot read the
                                results of running Blast with the parameter
                                '-outfmt' set to %s
                                """ % self.alignment_parameters["outfmt"])
        elif isinstance(file_format, list):
            raise TypeError("The result file for Blast cannot be a list.")
        else:
            bool_found = 0
            for k, v in BlastNAligner.outfmt_to_resultfile.items():
                if v == BlastNAligner.outfmt_to_resultfile[file_format]:
                    if k == file_format:
                        bool_found = 1
                        self.result_file = v
                        break
                elif v == file_format:
                    if k == file_format:
                        bool_found = 1
                        break
                    else:
                        raise ValueError("""The result file selected has been
                                         %s, while the 'outfmt' option %s (
                                         result file %s) has been runned"""
                                         % (self.result_file,
                                            file_format,
                                            BlastNAligner.outfmt_to_resultfile[
                                                self.alignment_parameters["outfmt"]]
                                            ))
            if bool_found == 0:
                raise ValueError("""The result file %s selected is not part of
                                    the options available (TSV, XML, TSV_extra,
                                    CSV)""" % self.result_file)

    def fit_alignment(self):

        self.get_resultfiles()

        header = BlastNAligner.get_header_format(
                    alignment_parameters=self.alignment_parameters)

        alignment = BlastNAlignment(output_path=None,
                                    filenames=self.alignment_files,
                                    result_file=self.result_file,
                                    headers=header)
        self.alignment = alignment

    @staticmethod
    def _create_extension(parameters):
        format_outfmt, _ = BlastNAligner.get_output_format(
                                    alignment_parameters=parameters)
        if format_outfmt == 5:
            extension = ".xml"
        elif format_outfmt == 10:
            extension = ".csv"
        else:
            extension = ".tsv"
        return extension



    def __call__(self, values_iter, variable_iter="subject"):

        commands = self.create_commands(parameters=self.alignment_parameters,
                                        output_param="output",
                                        variable_iter=variable_iter,
                                        values_iter=values_iter)
        super(BlastNAligner, self).__call__(commands)

        return self.result_dict["software_executions"]



class KMAAligner(_Aligner):

    def __init__(self, result_file=None):

        _Aligner.__init__(self, aligner="kma",
                          commandline_exec=KMACommandline,
                          result_file=result_file)

    def get_resultfiles(self):

        if self.result_file is None:
            if "sparse" in self.alignment_parameters:
                self.result_file = ["Sparse"]
            else:
                self.result_file = ["Result", "Fragments", "Consensus",
                                    "Alignment", "Matrix", "Mapstat",
                                    "Frag_Raw", "VCF"]

        if("non_consensus" in self.alignment_parameters
           and "Consensus" in self.result_file):
            self.result_file.remove("Consensus")
        if("no_aln" in self.alignment_parameters
           and "Alignment" in self.result_file):
            self.result_file.remove("Alignment")
        if("no_frag" in self.alignment_parameters
           and "Fragments" in self.result_file):
            self.result_file.remove("Fragments")
        if("extra_files" in self.alignment_parameters
           and "Mapstat" in self.result_file):
            self.result_file.remove("Mapstat")
        if("vcf" in self.alignment_parameters
           and "VCF" in self.result_file):
            self.result_file.remove("VCF")
        if("matrix" in self.alignment_parameters
            and "Matrix" in self.result_file):
            self.result_file.remove("Matrix")
        if("best_maps" in self.alignment_parameters
           and "Frag_Raw" in self.result_file):
            self.result_file.remove("Frag_Raw")

    def fit_alignment(self):

        self.get_resultfiles()

        alignment = KMAAlignment(output_path=None,
                                 filenames=self.alignment_files,
                                 result_file=self.result_file)
        self.alignment = alignment


    def __call__(self, values_iter, variable_iter="template_db"):


        commands = self.create_commands(parameters=self.alignment_parameters,
                                        output_param="output",
                                        variable_iter=variable_iter,
                                        values_iter=values_iter)

        super(KMAAligner, self).__call__(commands)

        return self.result_dict["software_executions"]



# Created by Alfred Ferrer Florensa
"""Contains objects for reading alignment files"""

import gzip
import pandas as pd
from cgelib.utils.file_mixin import _Parse_File


class Iterator_ResFile(_Parse_File):
    """Create iterator for a .res file"""
    STD_header={
         "Template": "templateID",
         "Score": "conclave_score",
         "Expected": "evalue",
         "Template_length": "template_length",
         "Template_Identity": "template_identity",
         "Template_Coverage": "template_coverage",
         "Query_Identity": "query_identity",
         "Query_Coverage": "query_coverage",
         "Depth": "depth",
         "q_value": "q_value",
         "p_value": "p_value"
     }

    EXTENSION = ".res"

    def __init__(self, path):
        self.header = None

        _Parse_File.__init__(self, path)

    def __repr__(self):

        return "Iterator_ResFile(%s)" % self.path

    def __str__(self):

        return self.path

    def __iter__(self):

        return self

    def __next__(self):
        entry_line = True
        while entry_line:
            line = self.parse_line()
            if line == "":
                entry = None
                _Parse_File.close(self)
            elif line.startswith("#"):
                self.header = line.replace("#",
                                           "").rstrip().replace(" ",
                                                                "").split("\t")
            else:
                line_split = line.rstrip().replace(" ", "").split("\t")
                if len(line_split) != len(self.header):
                    raise IndexError("Length of line is not equal to"
                                     " header")

                hit = Iterator_ResFile.format_entry(self.header, line_split)
                entry_line = False
        return hit

    @staticmethod
    def format_entry(header, data):
        hit = {}
        if isinstance(data, dict):
            hit.update(data)
        elif isinstance(data, list):
            if len(data) != len(header):
                raise IndexError("Length of line (%s)is not equal to"
                                 " header (%s). Check if you have chosen "
                                 "the right separator." % (len(data),
                                                           len(header)
                                                           ))
            for n_feat in range(len(data)):
                value = data[n_feat]
                key = header[n_feat]
                hit[key] = value
        else:
            raise TypeError("Data has to be list or dictionary")
        return hit


class Iterator_MapstatFile(_Parse_File):

    STD_header = {
        "refSequence": "templateID",
        "readCount": "reads_mapped",
        "fragmentCount": "fragments_mapped",
        "mapScoreSum": "mapScoreSum",
        "refCoveredPositions": "template_coveredPos",
        "refConsensusSum": "template_consesusSum",
        "bpTotal": "bpTotal",
        "depthVariance": "depth_variance",
        "nucHighDepthVariance": "nucHigh_depth_variance",
        "depthMax": "depth_max",
        "snpSum": "snps",
        "insertSum": "insertions",
        "deletionSum": "deletions",
        "fragmentCountAln": "fragments_mapped_align",
        "readCountAln": "reads_mapped_align"
        }

    EXTENSION = ".mapstat"

    def __init__(self, path):
        self.header = None
        _Parse_File.__init__(self, path)

    def get_info(self):
        info = {}
        if not self.file.closed:
            file_open = open(self.path, 'r')
        else:
            file_open = self.file
        for line in file_open:
            if line.startswith("##"):
                split_line = line.replace("## ", "").rstrip().split("\t")
                if len(split_line) != 2:
                    raise TypeError("The information line of Mapstat does"
                                    " not follow the normal format (%)" %
                                    (line))
                info[split_line[0]] = split_line[1]
            else:
                break
        return info

    def __repr__(self):

        return "Iterator_MapstatFile(%s)" % self.path

    def __str__(self):

        return self.path

    def __iter__(self):

        return self

    def __next__(self):
        entry_line = True
        while entry_line:
            line = self.parse_line()
            if line == "":
                entry = None
                _Parse_File.close(self)
            elif line.startswith("## "):
                pass
            elif line.startswith("#"):
                self.header = line.replace("#",
                                           "").rstrip().replace(" ",
                                                                "").split("\t")
            else:
                line_split = line.rstrip().replace(" ", "").split("\t")
                if len(line_split) != len(self.header):
                    raise IndexError("Length of line is not equal to"
                                     " header")
                hit = Iterator_MapstatFile.format_entry(self, line_split)
                entry_line = False
        return hit

    @staticmethod
    def format_entry(self, data):

        hit = {}
        if isinstance(data, dict):
            hit.update(data)
        elif isinstance(data, list):
            if len(data) != len(self.header):
                raise IndexError("Length of line (%s)is not equal to"
                                 " header (%s). Check if you have chosen "
                                 "the right separator." % (len(data),
                                                           len(self.header)
                                                           ))
            for n_feat in range(len(data)):
                value = data[n_feat]
                key = self.header[n_feat]
                hit[key] = value
        else:
            raise TypeError("Data has to be list or dictionary")
        return hit


class Iterator_MatrixFile(_Parse_File):
    """Create iterator for a .matrix file"""

    STD_header={
        "templateID": "templateID",
        "matrix": "matrix"}

    EXTENSION = ".mat.gz"

    def __init__(self, path, is_gzip=True):
        self.header = ["ref_nucl", "A_nucl", "C_nucl", "G_nucl", "T_nucl",
                       "N_nucl", "null_nucl"]

        _Parse_File.__init__(self, path, is_gzip)

        self.gene = None
        self.template_df = []

    def __repr__(self):

        return "Iterator_MatrixFile(%s)" % self.path

    def __str__(self):

        return self.path

    def __iter__(self):

        return self

    def __next__(self):
        entry_line = True
        while entry_line:
            line = self.parse_line()
            if line.startswith("#"):
                if self.gene is not None:
                    gene = self.gene
                    entry = pd.DataFrame(self.template_df, columns=self.header)
                    entry_line = False
                self.gene = line.replace("#", "").rstrip()
                self.template_df = []
            elif line == "\n":
                continue
            elif line == "":
                if self.gene is not None:
                    gene = self.gene
                    entry = pd.DataFrame(self.template_df, columns=self.header)
                    entry_line = False
                else:
                    entry = None
                _Parse_File.close(self)
            else:
                line_split = line.rstrip().split("\t")
                if len(line_split) != len(self.header):
                    raise IndexError("Length of line is not equal to"
                                     " header")
                self.template_df.append(line_split)
        hit = Iterator_MatrixFile.format_entry(gene, entry)
        return hit

    @staticmethod
    def format_entry(gene, data):
        hit = {}
        hit["templateID"] = gene
        hit["matrix"] = data
        return hit


class Iterator_AlignmentFile(_Parse_File):
    """Create iterator for a .aln file"""

    STD_header={
            "templateID": "templateID",
            "query_seq": "query_aln",
            "template_seq": "template_aln",
            "alignment_seq": "aln_scheme"
        }
    EXTENSION = ".aln"

    def __init__(self, path, is_gzip=False):

        _Parse_File.__init__(self, path, is_gzip)
        self.gene = None
        self.alignment = None

    def __repr__(self):

        return "Iterator_AlignmentFile(%s)" % self.path

    def __str__(self):

        return self.path

    def __iter__(self):

        return self

    def __next__(self):
        entry_line = True
        alignment = None
        while entry_line:
            line = self.parse_line()
            if line.startswith("# "):
                if self.gene is not None:
                    gene = self.gene
                    alignment = self.alignment
                    entry_line = False
                self.gene = line.replace("# ", "").rstrip()
                self.alignment = {"template_seq": [], "alignment_seq": [],
                                  "query_seq": []}
            elif line.startswith("template:"):
                temp_seq = str(line.split("\t")[-1].rstrip())
                self.alignment["template_seq"].append(temp_seq)
            elif line.startswith("query:"):
                temp_seq = str(line.split("\t")[-1].rstrip())
                self.alignment["query_seq"].append(temp_seq)
            elif line == "\n":
                continue
            elif line == "":
                # This might be useless
                if self.gene is not None:
                    entry = self.alignment
                    entry_line = False
                    break
                else:
                    entry = None
            else:
                aln_seq = str(line.split("\t")[-1].rstrip())
                self.alignment["alignment_seq"].append(aln_seq)
        if alignment is None:
            alignment = self.alignment
            gene = self.gene
            _Parse_File.close(self)
        hit = Iterator_AlignmentFile.format_entry(self, alignment, gene)
        return hit

    @staticmethod
    def format_entry(self, alignment, template):
        hit = {}
        hit["templateID"] = template
        hit["query_seq"] = "".join(alignment["query_seq"])
        hit["template_seq"] = "".join(alignment["template_seq"])
        hit["alignment_seq"] = "".join(alignment["alignment_seq"])
        return hit


class Iterator_ConsensusFile(_Parse_File):
    """Create iterator for a .fsa file"""

    STD_header={
     "templateID": "templateID",
     "sequence": "query_aln"
     }

    EXTENSION = ".fsa"

    def __init__(self, path):
        self.gene = None
        self.sequence = None

        _Parse_File.__init__(self, path)

    def __repr__(self):

        return "Iterator_ConsensusFile(%s)" % self.path

    def __str__(self):

        return self.path

    def __iter__(self):

        return self

    def __next__(self):
        entry_line = True
        entry = None
        while entry_line:
            line = self.parse_line()
            if line.startswith(">"):
                if self.gene is not None:
                    gene = self.gene
                    entry = "".join(self.sequence)
                    entry_line = False
                self.gene = line.replace(">", "").rstrip()
                self.sequence = []
            elif line == "\n":
                continue
            elif line == "":
                if self.gene is not None:
                    gene = self.gene
                    entry = "".join(self.sequence)
                    self.sequence = []
                    entry_line = False
                else:
                    entry = None    #BUG?
                _Parse_File.close(self)
            else:
                self.sequence.append(line.rstrip())
        if entry is None:
            entry = "".join(self.sequence)
            gene = self.gene
        hit = Iterator_ConsensusFile.format_entry(self, entry, gene)
        return hit

    @staticmethod
    def format_entry(self, sequence, template):
        hit = {}
        hit["templateID"] = template
        hit["query_aln"] = sequence
        return hit


class Iterator_VCFFile(_Parse_File):

    STD_header={
         "templateID": "templateID",
         "vcf_data": "point_variations"
         }
    EXTENSION = ".vcf.gz"

    def __init__(self, path, is_gzip=True):

        self.gene = None
        self.header = None
        self.template_df = []


        _Parse_File.__init__(self, path, is_gzip)

    def __repr__(self):

        return "Iterator_VCFFile(%s)" % self.path

    def __str__(self):

        return self.path

    def __iter__(self):

        return self

    def __next__(self):
        entry_line = True
        while entry_line:
            line = self.parse_line()
            if line.startswith("##"):
                pass
            elif line.startswith("#"):
                self.define_header(header_line=line)
            elif line == "":
                if self.gene is not None:
                    gene = self.gene
                    entry = pd.DataFrame(self.template_df, columns=self.header)
                    entry_line = False
                else:
                    entry = None
                _Parse_File.close(self)
            else:
                line_split = line.rstrip().split("\t")
                if self.gene == line_split[0]:
                    if len(line_split) != len(self.header):
                        raise IndexError("Length of line is not equal to"
                                         " header")
                    self.template_df.append(line_split)
                else:
                    if self.gene is not None:
                        gene = self.gene
                        entry = pd.DataFrame(self.template_df, columns=self.header)
                        entry_line = False
                    self.gene = line_split[0]
                    self.template_df = []
        hit = Iterator_VCFFile.format_entry(entry, gene)
        return hit

    def define_header(self, header_line):
        header_split = header_line.replace("#", "").rstrip().split("\t")
        last_col = header_split[-1]
        if len(header_split) != 10:
            raise IndexError("The iterator needs a vcf file with ten columns")
        else:
            header = ["templateID", "position", "id", "ref_base", "alt_base",
                      "quality", "filter", "info", "format", last_col]
        self.header = header


    def get_info(self):
        info = {}
        with gzip.open(self.path, 'rt') as file_open:
            for line in file_open:
                if line.startswith("##"):
                    line_split = line.replace("##", "").rstrip().split("=")
                    if len(line_split) == 2:
                        info[line_split[0]] = line_split[1]
                    else:
                        if line_split[0] not in info:
                            info[line_split[0]] = {}
                        new_value = []
                        for i in range(len(line_split[1:])-1):
                            new_element = line_split[1:][i].replace('<',
                                                                    '').split(
                                                                    ",")
                            new_value.extend(new_element)
                        new_value.append(line_split[-1].replace('>', ''))
                        dict_value = dict(zip(new_value[::2], new_value[1::2]))
                        info[line_split[0]].update({dict_value["ID"]:
                                                    dict_value})
                else:
                    break
        return info

    @staticmethod
    def format_entry(entry, gene):
        hit = {}
        hit["templateID"] = gene
        hit["vcf_data"] = entry
        return hit


class Iterator_SPAFile(_Parse_File):
    """Create iterator for .spa file"""

    STD_header={
        "#Template": "templateID",
        "Num": "Num",
        "Score": "score",
        "Expected": "evalue",
        "Template_length": "template_length",
        "Query_Coverge": "query_coverage",
        "Template_Coverage": "template_coverage",
        "Depth": "depth",
        "tot_query_Coverage": "tot_query_coverage",
        "tot_template_Coverage": "tot_template_coverage",
        "tot_depth": "tot_depth",
        "q_value": "q_value",
        "p_value": "p_value"
    }

    EXTENSION = ".spa"

    def __init__(self, path):
        self.header = None

        _Parse_File.__init__(self, path)

    def __repr__(self):

        return "Iterator_SPAFile(%s)" % self.path

    def __str__(self):

        return self.path

    def __iter__(self):

        return self

    def __next__(self):
        entry_line = True
        while entry_line:
            line = self.parse_line()
            if line == "":
                _Parse_File.close(self)
            elif line.startswith("#"):
                self.header=Iterator_SPAFile.define_header(header_str=line)
            else:
                line_split = line.rstrip().replace(" ", "").split("\t")
                if len(line_split) != len(self.header):
                    raise IndexError("Length of line is not equal to"
                                     " header")
                hit = Iterator_SPAFile.format_entry(self.header, line_split)
                entry_line = False
        return hit

    @staticmethod
    def define_header(header_str):
        header = header_str.replace("#", "").rstrip().replace(" ", "").split("\t")
        if header != Iterator_SPAFile.STD_header:
            raise ValueError("""Header of the %s file is not equal to the header
                                expected. Please inform curator.""")
        return header

    @staticmethod
    def format_entry(header, data):
        hit = {}
        if isinstance(data, dict):
            hit.update(data)
        elif isinstance(data, list):
            if len(data) != len(header):
                raise IndexError("Length of line (%s)is not equal to"
                                 " header (%s). Check if you have chosen "
                                 "the right separator." % (len(data),
                                                           len(header)
                                                           ))
            for n_feat in range(len(data)):
                value = data[n_feat]
                key = header[n_feat]
                hit[key] = value
        else:
            raise TypeError("Data has to be list or dictionary")


class Iterator_FragmentFile(_Parse_File):
    """Create iterator for .frag file"""
    STD_header={
            "templateID": "templateID",
            "fragments": "reads_aligned"
             }
    EXTENSION = ".frag.gz"
    def __init__(self, path, gzip=True):
        self.header = None
        self.gene = None
        self.template_df = []

        _Parse_File.__init__(self, path, gzip)

    def __repr__(self):

        return "Iterator_FragmentFile(%s)" % self.path

    def __str__(self):

        return self.path

    def __iter__(self):

        return self

    def __next__(self):
        entry_line = True
        while entry_line:
            line = self.parse_line()
            if line == "":
                if self.gene is not None:
                    gene = self.gene
                    entry = pd.DataFrame(self.template_df, columns=self.header)
                    entry_line = False
                else:
                    entry = None
                _Parse_File.close(self)
            else:
                line_split = line.rstrip().split("\t")
                if self.header is None:
                    self.define_header(line_split=line_split)
                else:
                    if len(line_split) != len(self.header):
                        raise IndexError("Length of line is not equal to"
                                         " header")
                if self.gene is None:
                    self.gene = line_split[5]
                    self.template_df = []

                if self.gene == line_split[5]:
                    #row_series = pd.Series(data=line_split, index=self.header)
                    self.template_df.append(line_split)
                else:
                    gene = self.gene
                    entry = pd.DataFrame(self.template_df, columns=self.header)
                    entry_line = False
                    self.gene = line_split[5]
                    self.template_df = []
                    self.template_df.append(line_split)


        hit = Iterator_FragmentFile.format_entry(entry, gene)
        return hit

    def define_header(self, line_split):
        if len(line_split) == 7:
            self.header = ["query_seq", "eq_mapped", "aln_score", "start_aln",
                           "end_aln", "template_name", "query_name"]
        elif len(line_split) == 9:
            self.header = ["query_seq", "eq_mapped", "aln_score", "start_aln",
                           "end_aln", "template_name", "query_name", "cut1",
                           "cut2"]
        else:
            raise KeyError("Fragment file is does has not 7 or 9 columns")

    @staticmethod
    def format_entry(data, gene):
        hit = {}
        hit["templateID"] = gene
        hit["fragments"] = data
        return hit



# Created by Alfred Ferrer Florensa
"""Contains objects for reading alignment files"""

import gzip
import io
import pandas as pd
import signal
import sys
from cgelib.sequence.SeqHit import AlnHit, BlastnHit
from cgelib.utils.file_mixin import _Parse_File, CGELibFileParseError
from Bio.Blast.NCBIXML import parse as BioXMLParse

class Iterator_XMLFile(_Parse_File):

    STD_header = {
        "queryID": "queryID",
        "templateID": "templateID",
        "query": "query_aln",
        "sbjct": "template_aln",
        "pident": "query_identity",
        "sbjct_start": "template_start_aln",
        "sbjct_end": "template_end_aln",
        "expect": "evalue",
        "align_length": "aln_length",
        "qcov": "query_coverage",
        "query_start": "query_start_aln",
        "query_end": "query_end_aln",
        "bits": "bitscore",
        "score": "score",
        "num_alignments": "n_alignments",
        "match": "aln_scheme",
        "identities": "n_identity",
        "positives": "n_positives",
        "gaps": "gaps",
        "strand": "strand",
        "frame": "frame"
    }

    EXTENSION = None

    def __init__(self, path):

        _Parse_File.__init__(self, path)
        self.iterator_XML = BioXMLParse(self.file)
        self.extension = "xml"
        self.recordblast = None
        self.recordaln = None
        self.recordhsp = None
        self.queryID = None


    def __iter__(self):
        return self

    def __next__(self):
        new_line = True
        while new_line:
            if self.recordblast is None:
                recordblast = self.parse_line(file_type="xml")
                self.recordblast = iter(recordblast.__dict__["alignments"])
                self.queryID = recordblast.__dict__["query"]
            if self.recordblast:
                if self.recordaln is None:
                    try:
                        self.recordaln = next(self.recordblast)
                        self.recordhsp = iter(self.recordaln.__dict__["hsps"])
                    except StopIteration:
                        self.recordblast = None
                        self.recordhsp = None
                        self.recordblast = None
                        #new_line=False
                if self.recordhsp:
                    try:
                        hsp_data = next(self.recordhsp)
                        new_line=False
                        hit = Iterator_XMLFile.format_entry(
                                                    queryID=self.queryID,
                                                    aln_info=self.recordaln,
                                                    hsp_info=hsp_data)
                        return hit
                    except StopIteration:
                        self.recordaln = None
                        self.recordhsp = None



    @staticmethod
    def format_entry(queryID, aln_info, hsp_info):
        hit = {}
        hit["queryID"] = queryID
        aln_dict = aln_info.__dict__
        hit["gene_length"] = aln_dict["length"]
        hit["templateID"] = aln_dict["hit_def"]
        hsp_dict = hsp_info.__dict__
        hit.update(hsp_dict)
        return hit




class Iterator_BlastSepFile(_Parse_File):

    STD_header = {
        "qseqid": "queryID",
        "qgi": "queryGI",
        "qacc": "query_accuracy",
        "sseqid": "templateID",
        "sallseqid": "templateIDs",
        "sgi": "templateGI",
        "sallgi": "templateGIs",
        "sacc": "template_accuryacy",
        "sallacc": "template_accuracies",
        "qstart": "query_start",
        "qend": "query_end",
        "pident": "query_identity",
        "sstart": "template_start_aln",
        "send": "template_end_aln",
        "qseq": "query_aln",
        "sseq": "template_aln",
        "evalue": "evalue",
        "length": "aln_length",
        "qstart": "query_start_aln",
        "qend": "query_end_aln",
        "bitscore": "bitscore",
        "score": "score",
        "nident": "n_identity",
        "positive": "n_positives",
        "gapopen": "gaps_open",
        "gaps": "gaps",
        "ppos": "positives",
        "frames": "frames",
        "qframe": "query_frame",
        "sframe": "template_frame",
        "btop": "btop",
        "staxids": "template_taxids",
        "sscinames": "template_sciname",
        "scomnames": "template_comname",
        "sblastnames": "template_blastname",
        "sskingdoms": "template_superkingdom",
        "stitle": "template_title",
        "salltitles": "template_titles",
        "sstrand": "template_strand",
        "qcovs": "query_coverage_template",
        "qcovhsp": "query_coverage_hsp",
        "qcovus": "query_coverage_once",
        "mismatch": "mismatch",
        }

    EXTENSION = None

    def __init__(self, path, separator="tab", comment_lines=False,
                 header=["qseqid", "sseqid", "pident", "length", "mismatch",
                         "gapopen", "qstart", "qend", "sstart", "send",
                         "evalue", "bitscore"]):
        self.header = header
        self.software = "blastn"
        if separator == "comma":
            self.separator = ","
            self.file_type = "csv"
        elif separator == "tab":
            self.separator = "\t"
            self.file_type = "tsv"
        else:
            raise ValueError("The separator of the blast results file has to "
                             "be 'comma' or 'tab'.")

        if separator == "comma" and comment_lines:
            raise TypeError("The option of comment_lines is only available "
                            "for tab separated files")
        self.comment_lines = comment_lines

        _Parse_File.__init__(self, path)

    def __repr__(self):

        return "Iterator_TabFile(%s)" % self.path

    def __str__(self):

        return self.path

    def __iter__(self):

        return self

    def __next__(self):
        entry_line = True
        while entry_line:
            line = self.parse_line()
            if line == "":
                entry = None
                _Parse_File.close(self)
                raise CGELibFileParseError(
                    "The iterator has arrived to the end of the file")
            elif self.comment_lines and line.startswith("#"):
                line_split = line.rstrip().split(" ")
                if line_split[1] == "Query:":
                    query_entry = line_split[-1]
                elif line_split[1] == "Subject:":
                    subject_entry = line_split[-1]
                elif line_split[-1] == "found":
                    if line_split[1] == "0":
                        pass
                        #print(query_entry, line_split)
                        #entry = {"qseqid": query_entry,
                        #         "templateID": subject_entry}
                        #hit = Iterator_BlastSepFile.format_entry(
                        #                        header=self.header, data=entry)
                        #return hit
                    else:
                        hits_found = int(line_split[1])
                elif line_split[1] == "Fields:":
                    line_header = line.replace("# Fields: ", "")
                    header = line_header.split(",")
            else:
                line_split = line.rstrip().replace(" ",
                                                   "").split(self.separator)
                if len(line_split) != len(self.header):
                    raise IndexError("Length of line '%s' (%s) is not equal to"
                                     " header '%s' (%s). Check if you have "
                                     "chosen the right separator." % (
                                        line_split, len(line_split),
                                        self.header, len(self.header)))
                hit = Iterator_BlastSepFile.format_entry(
                                header=self.header,data=line_split)
                entry_line = False
        return hit

    @staticmethod
    def format_entry(header, data):
        hit = {}
        if isinstance(data, dict):
            hit.update(data)
            hit["empty"] = True
        elif isinstance(data, list):
            if len(data) != len(header):
                raise IndexError("Length of line (%s)is not equal to"
                                 " header (%s). Check if you have chosen "
                                 "the right separator." % (len(data),
                                                           len(header)
                                                           ))
            for n_feat in range(len(data)):
                hit[header[n_feat]] = data[n_feat]
            hit["empty"] = False
        else:
            raise TypeError("Data has to be list or dictionary")
        return hit



    @staticmethod
    def assign_BlastnHit(self, data, empty=False):
        hit = BlastnHit(empty=empty, file_type=self.file_type)
        if isinstance(data, dict):
            for key, value in data.items():
                key = Translate.translate_keys(key,
                                                    Iterator_BlastSepFile.Translator_Hit)
                hit[key] = value
        elif isinstance(data, list):
            if len(data) != len(self.header):
                raise IndexError("Length of line (%s)is not equal to"
                                 " header (%s). Check if you have chosen "
                                 "the right separator." % (len(data),
                                                           len(self.header)
                                                           ))
            for n_feat in range(len(data)):
                value = data[n_feat]
                key = Translate.translate_keys(self.header[n_feat],
                                                    Iterator_BlastSepFile.Translator_Hit)
                hit[key] = value
        else:
            raise TypeError("Data has to be list or dictionary")

        return hit



#!/usr/bin/env python3


class LoadersMixin():

    @staticmethod
    def load_md_table_after_keyword(mdfile, keyword, header_key=0,
                                    ignore_case=True):
        md_table_list = []
        record_line = False

        with open(mdfile, "r") as fh:
            for line in fh:

                if(line.startswith(keyword)):
                    record_line = True

                elif(record_line and line.strip()):
                    if(not line.startswith("|")):
                        record_line = False
                    else:
                        md_table_list.append(line.strip())

        md_table_str = "\n".join(md_table_list)

        return LoadersMixin.load_md_table_to_dict(md_table_str, header_key,
                                                  ignore_case)

    @staticmethod
    def load_md_table_to_dict(md_str, key=0, ignore_case=True):
        """ Input:
                md_str: String containing markdown table
                key: Integer or string. If integer, use the column at that
                     index as keys. If string then the column header matching
                     the string will be used as key.
            Output:
                out_dict: Dictonary containing table information. Keys are as
                specified under 'key' argument. A value is a list containing
                the row information. The key will not be part of the row
                information. Headers of the table will be stored under the
                key 'headers'.

            Loads a markdown formatted string into a dict object, with specified
            column as keys and the remaining row information stored as a list.

        """
        md_str = md_str.strip()
        md_list = md_str.split("\n")

        headers = LoadersMixin.split_and_clean_str(md_list[0], "|")
        header_index = LoadersMixin._get_header_index(key, headers, ignore_case)

        out_dict = {}

        # Skip header and horisontal seperator
        for line in md_list[2:]:

            entries = LoadersMixin.split_and_clean_str(line, "|")

            val_list = []
            for i, entry in enumerate(entries):
                if(i == header_index):
                    entry_key = entry
                else:
                    val_list.append(entry)

            out_dict[entry_key] = val_list

        return out_dict

    @staticmethod
    def _get_header_index(key, headers, ignore_case):
        if(isinstance(key, int)):
            index = key
        else:
            index = 0
            for header in headers:
                if(ignore_case and header.lower() == key.lower()):
                    break
                elif(header == key):
                    break
                else:
                    index = index + 1

        if(index >= len(headers) and isinstance(key, str)):
            raise KeyError("The argument provided as key: {} did not match any"
                           " of the strings in the list of headers: {}"
                           .format(key, headers))
        if(index >= len(headers) and isinstance(key, int)):
            raise IndexError("The argument provided as key: {} was larger or "
                             "equal to the number of headers: {}."
                             .format(key, len(headers)))

        else:
            return index

    @staticmethod
    def split_and_clean_str(string, delimiter):
        """ Input:
                string: String to split and clean.
                delimiter: Delimiter used for split.
            Output:
                tmp_lst: List without empty entries and without leading and
                         trailing whitespaces in entries.

        """
        tmp_lst = [x.strip() for x in string.split(delimiter)]
        return list(filter(None, tmp_lst))



# Created by Alfred Ferrer Florensa
"""Contains objects to contain HIT"""

import os
import sys
import json
import pandas as pd
from cgelib.output.result import Result


class CGEFeatureValueError(ValueError):
    """ Error related to a ValueError of the value of the feature
    """

    def __init__(self, message, *args):
        self.message = message
        # allow users initialize misc. arguments as any other builtin Error
        super(CGEFeatureValueError, self).__init__(message, *args)


class CGEFeatureFormatError(TypeError):
    """ Error related to a TypeError of the value of the feature
    """

    def __init__(self, message, *args):
        self.message = message
        # allow users initialize misc. arguments as any other builtin Error
        super(CGEFeatureFormatError, self).__init__(message, *args)


class CGEFeatureKeyError(KeyError):
    """ Error related to a KeyError of the value of the feature
    """

    def __init__(self, message, *args):
        self.message = message
        # allow users initialize misc. arguments as any other builtin Error
        super(CGEFeatureKeyError, self).__init__(message, *args)


class AlnFeature:

    @staticmethod
    def translate_entry(file_iterator, entry):
        dictionary = file_iterator.STD_header
        file_name = file_iterator.__name__.split("_")[1]
        new_entry = {}
        for k in entry:
            if k in dictionary:
                new_entry[dictionary[k]] = entry[k]
            else:
                provisional_key = "{}_{}_undescribed".format(k, file_name)
                new_entry[provisional_key] = entry[k]
        return new_entry

    @staticmethod
    def FeatureDF_to_dict(feature_DF, aln_hit_key, resultfile):
        feature = {}
        for index, row in feature_DF.iterrows():
            feature_entry = {}
            if resultfile == "Matrix":
                feature_entry["type"] = "matrix_position"
                feature_entry["position"] = index
                feature_entry["key"] = "{}_Pos{}".format(aln_hit_key, index)
            elif resultfile == "Fragments":
                feature_entry["type"] = "fragment_aligned"
                feature_entry["key"] = "{}_Frag{}".format(aln_hit_key, index)
            elif resultfile == "VCF":
                feature_entry["type"] = "point_variation"
                feature_entry["position"] = index
                feature_entry["key"] = "{}_Variation{}".format(aln_hit_key,
                                                               index)
            feature_entry.update(row.to_dict())
            feature.update({feature_entry["key"]: feature_entry})
        return feature

    @staticmethod
    def KMAMatrix_to_std(mat_DF, aln_hit_key):
        for index, row in mat_DF.iterrows():
            matrix_entry = {}
            matrix_entry["type"] = "matrix_position"
            matrix_entry["position"] = index
            matrix_entry["key"] = "{}_Pos{}".format(aln_hit_key, index)
            matrix_entry["ref_nucleotide"] = row["Nucleotide"]
            matrix_entry["A_nucleotide"] = row["A"]
            matrix_entry["G_nucleotide"] = row["G"]
            matrix_entry["T_nucleotide"] = row["T"]
            matrix_entry["C_nucleotide"] = row["C"]
            matrix_entry["N_nucleotide"] = row["N"]
            matrix_entry["null_nucleotide"] = row["-"]
            yield matrix_entry

    @staticmethod
    def KMAFrag_to_std(frag_DF):
        for index, row in frag_DF.iterrows():
            matrix_entry = {}
            matrix_entry["type"] = "fragment_aligned"
            matrix_entry["key"] = "{}_Frag{}".format(aln_hit_key, index)
            matrix_entry["query_seq"] = row["query_seq"]
            matrix_entry["equal_mapped_templ"] = row["eq_mapped"]
            matrix_entry["alignment_score"] = row["aln_score"]
            matrix_entry["start_aln"] = row["start_aln"]
            matrix_entry["end_aln"] = row["end_aln"]
            matrix_entry["template_name"] = row["template"]
            matrix_entry["query_name"] = row["query_name"]
            if "cut_start" in row and "cut_end" in row:
                matrix_entry["cut_start"] = row["cut_1"]
                matrix_entry["cut_end"] = row["cut_2"]
            yield matrix_entry

    @staticmethod
    def KMAVCF_to_std(vcf_DF, aln_hit_key, hit_file):
        "CHROM  POS ID REF ALT QUAL FILTER  INFO  FORMAT         fosfomycin"
        for index, row in vcf_DF.iterrows():
            matrix_entry = {}
            matrix_entry["type"] = "point_variation"
            matrix_entry["position"] = index
            matrix_entry["key"] = "{}_Variation{}".format(aln_hit_key, index)
            matrix_entry["id"] = row["ID"]
            matrix_entry["reference_base"] = row["REF"]
            matrix_entry["alternative_base"] = row["ALT"]
            matrix_entry["quality"] = row["QUAL"]
            matrix_entry["filter"] = row["FILTER"]
            matrix_entry["information"] = row["INFO"]
            matrix_entry["format"] = row[str(hit_file)]
            yield matrix_entry



#!/usr/bin/env python3

import subprocess
from git import Repo
from git.exc import InvalidGitRepositoryError


class PliersMixin():

    DB_VERSION_FILE = "VERSION"

    @staticmethod
    def get_version_commit(gitdir):
        """
            Input: Path to git directory
            Return: (version, commmit_hash)

            commmit_hash: The 40 character hash describing the exact commit of
            the git directory.
            version: The tag of the current commit. If no tag exists the first
            7 characters of the commit hash will be returned instead.

            If the given directory is not a git directory the returned values
            will be 'unknown'.
            Likewise the git module may raise a ValueError if there are
            problems with permissions in a git directory. In this case the
            function will also return 'unknown'.
        """
        try:
            repo = Repo(gitdir)

            com2tag = {}
            for tag in repo.tags:
                com2tag[tag.commit.hexsha] = str(tag)

            version = com2tag.get(repo.commit().hexsha, repo.commit().hexsha[:7])

            return (version, repo.commit().hexsha)
        except (InvalidGitRepositoryError, ValueError):
            return ("unknown", "unknown")

    @staticmethod
    def get_version_pymodule(name):
        """
            Input: python module name
            Return: (version)

            version: The version of the module.

            If the given directory is not a python module it will return
            'unknown'
        """
        try:
            from importlib.metadata import version
            version = version(str(name))
        except ModuleNotFoundError:
            try:
                from pkg_resources import get_distribution
                from pkg_resources import DistributionNotFound
                version = get_distribution(str(name)).version
            except DistributionNotFound or AttributeError:
                try:
                    from importlib import import_module
                    mymodule = import_module(str(name))
                    version = mymodule.__version__
                except ModuleNotFoundError or NameError:
                    version = "unknown"
        return version

    @staticmethod
    def get_version_database(db_path):
        try:
            with open(f"{db_path}/{PliersMixin.DB_VERSION_FILE}") as fh:
                version = fh.readline().strip()
        except FileNotFoundError:
            version = "unknown"
        return version

    @staticmethod
    def get_git_tag(folders_path):
        try:
            git_tag = str(
                subprocess.check_output(
                    ['git', 'describe', '--tags', '--abbrev=0'],
                    stderr=subprocess.STDOUT, cwd=folders_path,
                )).strip('\'b\\n')
        except subprocess.CalledProcessError as exc_info:
            match = re.search(
                'fatal: no tag exactly matches \'(?P<commit>[a-z0-9]+)\'', str(exc_info.output))
            if match:
                raise Exception(
                    'Bailing: there is no git tag for the current commit, {commit}'.format(
                    commit=match.group('commit')))
            raise Exception(str(exc_info.output))
        return git_tag



from setuptools import setup, find_packages
import site
import sys
site.ENABLE_USER_SITE = "--user" in sys.argv[1:]
setup(
    name='cgelib',
    version='0.7.5',
    url='https://bitbucket.org/genomicepidemiology/cgelib.git',
    author='Center for Genomic Epidemiology',
    author_email='food-cgehelp@dtu.dk',
    description='Description of my package',
    packages=find_packages(),
    install_requires=['python-dateutil', 'GitPython'],
)



#!/usr/bin/env python3


class CGECoreOut(Exception):
    """ Root for all submodule out exceptions in the cgecore module.
        Only used to enable "except" of all exceptions from submodule.
        Never raised.
    """
    pass


class CGECoreOutTypeError(CGECoreOut):
    """ Raised when the 'type' specified is not defined in the json
        definition.
    """
    def __init__(self, message, *args):
        self.message = message
        # allow users initialize misc. arguments as any other builtin Error
        super(CGECoreOutTypeError, self).__init__(message, *args)


class CGECoreOutInputError(CGECoreOut):
    """ Raised when some of the data stored in a Result object did not pass
        validation based on the JSON definition.
    """
    def __init__(self, message, errors, *args):
        self.message = message
        self.errors = errors
        # allow users initialize misc. arguments as any other builtin Error
        super(CGECoreOutInputError, self).__init__(message, errors, *args)


class CGECoreOutTranslateError(CGECoreOut):
    """ Raised when keys for the given type of Translate object does not match"
        a key in the JSON definition
    """
    def __init__(self, message, *args):
        self.message = message
        # allow users initialize misc. arguments as any other builtin Error
        super(CGECoreOutTranslateError, self).__init__(message, *args)


class CGECoreOutDuplicateKeyError(CGECoreOut):
    """ Raised when an non-empty entry in a dict would be overwritten be
        overwritten by an new entry with an identical key.
    """
    def __init__(self, message, *args):
        self.message = message
        # allow users initialize misc. arguments as any other builtin Error
        super(CGECoreOutDuplicateKeyError, self).__init__(message, *args)



#!/usr/bin/env python3

import dateutil.parser
import pandas as pd

class ValueParsers():

    @staticmethod
    def parse_bool(val):
        """
        Test if value is either of the python bool types or if it is one of the
        strings: True or False. The string test is not case sensitive.
        """
        accepted_vals = ("true", "false")
        if(val is True or val is False):
            return
        elif(str(val).lower() in accepted_vals):
            return
        else:
            return ("Value must be a boolean (True/False). Value was: {}"
                    .format(val))

    @staticmethod
    def parse_bool_or_unknown(val):
        """
        Test if value is either of the python bool types or if it is one of the
        strings: True, False, or Unknown. The string test is not case sensitive.
        """
        accepted_vals = ("true", "false", "unknown")
        if(val is True or val is False):
            return
        elif(str(val).lower() in accepted_vals):
            return
        else:
            return ("Value must be a boolean (True/False) or unknown. Value "
                    "was: {}".format(val))

    @staticmethod
    def parse_char64(val):
        """
        Test val is exactely 64 characters. Can be anything that can be handled
        by str()
        """
        val = str(val)
        if(len(val) != 64):
            return ("This field expects a string of lenght 64 but the lenght "
                    "of the string is {}. The string is: {}"
                    .format(len(val), val))

    @staticmethod
    def parse_date(val):
        """
        Test str(val) can be converted using dateutil.parser.isoparse()
        """
        try:
            # If the date is just a year it might be an integer (ex. 2018)
            dateutil.parser.isoparse(str(val))
        except ValueError:
            return ("Date format not recognised. Date format must adhere to "
                    "the ISO 8601 format (YYYY-MM-DD). Provided value was: {}"
                    .format(val))

    @staticmethod
    def parse_integer(val):
        """
        Test that val is an integer
        """
        try:
            val = int(float(val))
        except ValueError:
            return "Value must be an integer. Value was: {}".format(val)

    @staticmethod
    def parse_percentage(val):
        """
        Test that val is between 0 and 100.
        """
        try:
            val = float(val)
        except ValueError:
            return "Value must be a number. Value was: {}".format(val)
        if(val < 0 or val > 100):
            return ("Percentage value must be between 0 and 100. The value "
                    "was: {}".format(val))

    @staticmethod
    def parse_string(val):
        try:
            val = str(val)
        except ValueError:
            return "Value could not be converted to a string."

    @staticmethod
    def parse_float(val):
        try:
            val = float(val)
        except ValueError:
            return "Value must be a float. Value was: {}".format(val)

    @staticmethod
    def parse_dataframe(val):
        if isinstance(val, type(pd.DataFrame)):
            val = val
        else:
            return "Value must be a pandas DataFrame"

    @staticmethod
    def parse_dictionary(val):
        if isinstance(val, dict):
            val = val
        else:
            return "Value must be a dictionary"



#!/usr/bin/env python3

import logging
import json
import os.path
from datetime import datetime, timezone
from collections import UserDict

from cgelib.utils.pliers_mixin import PliersMixin
from cgelib.utils.transformer_mixin import TransformerMixin
from cgelib.output.parserdict import ParserDict
from cgelib.output.exceptions import (CGECoreOutTypeError, CGECoreOutInputError,
                                      CGECoreOutDuplicateKeyError)


class Result(UserDict):

    BEONE_JSON_FILE = "beone/beone.json"
    TEMPLATE_DIR = "templates_json"
    beone_json_path = os.path.join(os.path.dirname(__file__), TEMPLATE_DIR,
                                   BEONE_JSON_FILE)

    def __init__(self, type, fmt_file=beone_json_path,
                 parsers=None, **kwargs):
        super().__init__()

        self.defs = {}
        # classes in a template translates to Result objects.
        self.classes = set()
        self.errors = {}
        with open(fmt_file, "r") as fh:
            self.defs = json.load(fh)

        if(parsers is None):
            self.val_parsers = ParserDict()
        else:
            self.val_parsers = ParserDict(parsers)

        self._set_type(type)
        self._parser = ResultParser(result_def=self.defs[type])
        for d in self._parser.arrays:
            self[d] = []
        for d in self._parser.dicts:
            self[d] = {}

        self.add(**kwargs)

    @staticmethod
    def init_software_result(name, gitdir, type="git_repo"):
        """
            Input: software_name, path to git directory
            Return: Result object with type: software_result
        """
        if(type == "python_module"):
            version = PliersMixin.get_version_pymodule(name)
            commit = "unknown"
        elif(type == "git_repo"):
            version, commit = PliersMixin.get_version_commit(gitdir)
        else:
            version, commit = ("unknown", "unknown")
        date = datetime.now(timezone.utc).date().isoformat()

        result_dict = {
            "type": "software_result",
            "software_name": name,
            "software_version": version,
            "software_commit": commit,
            "run_date": date,
            "key": "{}-{}".format(name, version)
        }
        return Result(**result_dict)

    def init_database(self, name, db_dir):
        database_metadata = {}
        database_metadata["type"] = "database"
        database_metadata["database_name"] = name

        if hasattr(PliersMixin, 'get_version_database'):
            version = PliersMixin.get_version_database(db_dir)
        else:
            version = 'unknown'
        version_in_git, commit = PliersMixin.get_version_commit(db_dir)

        if version == 'unknown':
            version = version_in_git

        database_metadata["database_version"] = version
        database_metadata["key"] = "{}-{}".format(name, version)
        database_metadata["database_commit"] = commit

        self.add_class(cl="databases", **database_metadata)

    def get_db_key(self, name):
        """
            Input:
                name: database name for which key(s) are desired.
            Ouput:
                list of keys: List of all keys which values match the input.

            Method for retrieving all keys for a specific database name. Often
            you will only expect a list with a single entry.
        """
        key_list = []
        for key, val in self["databases"].items():
            if(val["database_name"] == name):
                key_list.append(key)
        return key_list

    def _set_type(self, type):
        if(type in self.defs):
            self["type"] = type
        else:
            raise CGECoreOutTypeError(
                "Unknown result type given. Type given: {}. Type must be one "
                "of:\n{}".format(type, list(self.defs.keys())))

    def add(self, **kwargs):
        for key, val in kwargs.items():
            if(val is None):
                continue
            self[key] = val

    def add_class(self, cl, type, clobber=True, clobber_warn=True, **kwargs):
        res = Result(type=type, **kwargs)
        self.classes.add(cl)
        if(cl in self._parser.arrays):
            self[cl].append(res)
        elif(cl in self._parser.dicts):
            self._add_class_dict(cl, res, clobber, clobber_warn)
            # self[cl][res["key"]] = res
        # Do not store the result object in neither a dict or a list.
        else:
            self[cl] = res

    def _add_class_dict(self, cl, res, clobber, clobber_warn):
        # val = self[cl][res["key"]]
        val = self[cl].get(res["key"], None)
        if(val):
            if(clobber):
                if(clobber_warn):
                    logging.warn("In {}. Clobbered entry with the key: {}."
                                 .format(cl, res["key"]))
            else:
                raise CGECoreOutDuplicateKeyError(
                    "The add_class would have clobbered an existing non-empty "
                    "entry. If this was intended you need to call the method "
                    "with the clobber argument set to True. Duplicate key: {}"
                    .format(res["key"])
                )
        self[cl][res["key"]] = res

    def check_results(self, strict=False, errors=None):
        """ Populates self.errors if any errors are encountered """

        for key, val in self.items():
            if(key == "type"):
                continue
            # The key is not defined
            elif(key not in self.defs[self["type"]]):
                if(strict):
                    self.errors[key] = ("Key not defined in template: {}"
                                        .format(key))
                    continue
                else:
                    continue
            self._check_result(key, val, self.errors)

        # errors is not None if called recursively
        if(errors is not None):
            errors[self["key"]] = self.errors
            return None
        # errors is None if it is the first/root call
        elif(errors is None and self._no_errors(self.errors)):
            return None
        else:
            raise CGECoreOutInputError(
                "Some input data did not pass validation, please consult the "
                "Dictionary of ERRORS:{}".format(self.errors),
                self.errors)

    def _check_result(self, key, val, errors, index=None):
        # Remember Result is a dict object and therefore this test should
        # be before the dict test.
        if(isinstance(val, Result)):
            val.check_results(errors)
        elif(isinstance(val, dict)):
            self._check_result_dict(key, val, errors)
        elif(isinstance(val, list)):
            self._check_result_list(key, val, errors)
        else:
            self._check_result_val(key, val, errors, index)

    def del_entries_by_values(self, values):
        values = tuple(values)
        deleted_keys = []
        for key, entry_val in self.items():
            if(key == "type"):
                continue
            if(entry_val in values):
                deleted_keys.append(key)
        for key in deleted_keys:
            del self[key]
        return deleted_keys

    def json_dumps(self):
        return json.dumps(TransformerMixin.userdict_to_dict(self))

    def _no_errors(self, errors):
        no_errors = True

        for key, val in errors.items():

            if(isinstance(val, dict)):
                no_errors = self._no_errors(val)
                if(no_errors is False):
                    return False

            elif(val is not None):
                return False

        return no_errors

    def _check_result_val(self, key, val, errors, index=None):
        val_type = self._parser[key]

        if(val_type.endswith("*")):
            val_type = val_type[:-1]

        val_error = self.val_parsers[val_type](val)

        if(val_error):
            if(index is not None):
                val_error = "{}:{} ".format(index, val_error)
            errors[key] = val_error

    def _check_result_dict(self, result_key, result_dict, errors):
        errors[result_key] = {}
        for key, val in result_dict.items():
            self._check_result(key, val, errors[result_key])

    def _check_result_list(self, result_key, result_list, errors):
        errors[result_key] = {}
        for i, val in enumerate(result_list):
            self._check_result(result_key, val, errors[result_key], index=i)


class ResultParser(dict):
    """"""
    def __init__(self, result_def):
        # self.classes = set()
        self.arrays = {}
        self.dicts = {}

        for key, val_def_str in result_def.items():
            val_def, *sub_def = val_def_str.split(" ")
            if(sub_def and val_def == "dict"):
                self.dicts[key] = sub_def[0]
                self[key] = sub_def[0]
            elif(sub_def and val_def == "array"):
                self.arrays[key] = sub_def[0]
                self[key] = sub_def[0]
            else:
                self[key] = val_def
